import unittest
import asyncio
from queue import Queue
from operator import itemgetter
from cryptos import *
from cryptos import coins_async
from cryptos.utils import alist
from cryptos.types import Tx, TxOut
from cryptos.electrumx_client.types import ElectrumXTx, ElectrumXMultiBalanceResponse, ElectrumXUnspentResponse
from typing import AsyncGenerator, Any, Union, List, Optional, Type


class BaseAsyncCoinTestCase(unittest.IsolatedAsyncioTestCase):
    name: str = ""
    coin: Type[coins_async.BaseCoin] = coins_async.Bitcoin
    addresses: List[str] = []
    segwit_addresses: List[str] = []
    native_segwit_addresses: List[str] = []
    multisig_addresses: List[str] = []
    cash_multisig_addresses: List[str] = []
    native_segwit_multisig_addresses: List[str] = []
    cash_addresses: List[str] = []

    fee: int = 500
    max_fee: int = 3500
    testnet: bool = True
    min_latest_height: int = 99999999999

    unspent_addresses: List[str] = []
    unspent: List[ElectrumXTx] = []
    unspents: List[ElectrumXTx] = []

    txheight: int = None
    privkeys: List[str] = []     # Private keys for above addresses in same order
    privkey_standard_wifs: List[str] = []
    privkey_segwit_wifs: List[str] = []
    privkey_native_segwit_wifs: List[str] = []
    txid: str = None
    txinputs: List[TxOut] = None
    block_hash: str = None

    balance: ElectrumXMultiBalanceResponse = {}
    balances: List[ElectrumXMultiBalanceResponse] = []
    history: List[ElectrumXTx] = {}
    histories: List[ElectrumXTx] = []
    raw_tx: str = ''
    expected_tx_verbose_keys: List[str] = [
        'blockhash', 'blocktime', 'confirmations', 'hash', 'hex', 'locktime', 'size', 'time', 'txid',
        'version', 'vin', 'vout']

    @classmethod
    def setUpClass(cls):
        print('Starting %s tests' % cls.name)

    def setUp(self) -> None:
        self._coin = self.coin(testnet=self.testnet, connection_timeout=100)

    async def asyncTearDown(self) -> None:
        await self._coin.close()

    def assertUnorderedListEqual(self, list1, list2, key):
        list1 = sorted(list1, key=itemgetter(key))
        list2 = sorted(list2, key=itemgetter(key))
        self.assertEqual(list1, list2)

    async def mock_electrumx_send_request(self, method: str, args=(),
                                          **kwargs) -> Optional[Union[float, ElectrumXUnspentResponse, str, List[str]]]:
        if method == "blockchain.scripthash.listunspent":
            scripthash = args[0]
            available_addresses = [addresses[0] for addresses in (
                self.addresses, self.segwit_addresses, self.native_segwit_addresses,
                self.multisig_addresses, self.cash_addresses, self.native_segwit_multisig_addresses,
                [privtopub(p) for p in self.privkeys]
            ) if addresses]
            if any(scripthash == self._coin.addrtoscripthash(address) for address in
                   available_addresses):
                return deepcopy(self.unspent)
            return []
        if method == "blockchain.transaction.broadcast":
            tx = args[0]
            tx_hash = public_txhash(
                tx
            )
            return tx_hash
        elif method == "server.version":
            return ["ElectrumX 1.16.0", "1.4"]
        elif method == "server.ping":
            return None
        elif method == "blockchain.estimatefee":
            return 1e-05

    @property
    def tx(self) -> Tx:
        return deserialize(self.raw_tx)

    def assertStandardWifOK(self):
        for privkey, expected_wif, address in zip(self.privkeys, self.privkey_standard_wifs, self.addresses):
            frmt = get_privkey_format(privkey)
            wif_format = "wif_compressed" if "compressed" in frmt else "wif"
            wif = self._coin.encode_privkey(privkey, formt=wif_format)
            self.assertEqual(wif, expected_wif)
            self.assertEqual(self._coin.privtop2pkh(privkey), address)
            self.assertEqual(self._coin.privtop2pkh(wif), address)
            self.assertEqual(self._coin.privtoaddr(privkey), address)
            self.assertEqual(self._coin.privtoaddr(wif), address)
            self.assertEqual(self._coin.pubtoaddr(privtopub(privkey)), address)
            self.assertEqual(self._coin.pubtoaddr(privtopub(wif)), address)
            self.assertTrue(self._coin.is_p2pkh(address))
            self.assertTrue(self._coin.is_address(address))
            self.assertFalse(self._coin.is_p2sh(address))
            if self._coin.segwit_supported:
                self.assertFalse(self._coin.is_native_segwit(address))

    def assertP2WPKH_P2SH_WifOK(self):
        for privkey, expected_wif, address in zip(self.privkeys, self.privkey_segwit_wifs, self.segwit_addresses):
            wif = self._coin.encode_privkey(privkey, formt="wif_compressed", script_type="p2wpkh-p2sh")
            self.assertEqual(wif, expected_wif)
            self.assertEqual(self._coin.privtop2wpkh_p2sh(privkey), address)
            self.assertEqual(self._coin.privtop2wpkh_p2sh(wif), address)
            self.assertEqual(self._coin.privtoaddr(wif), address)
            self.assertEqual(self._coin.pubtop2wpkh_p2sh(privtopub(privkey)), address)
            self.assertEqual(self._coin.pubtop2wpkh_p2sh(privtopub(wif)), address)
            self.assertTrue(self._coin.is_p2sh(address))
            self.assertTrue(self._coin.is_address(address))
            self.assertFalse(self._coin.is_p2pkh(address))
            self.assertFalse(self._coin.is_native_segwit(address))

    def assertP2WPKH_WIFOK(self):
        for privkey, expected_wif, address in zip(self.privkeys, self.privkey_native_segwit_wifs, self.native_segwit_addresses):
            wif = self._coin.encode_privkey(privkey, formt="wif_compressed", script_type="p2wpkh")
            self.assertEqual(wif, expected_wif)
            self.assertEqual(self._coin.privtosegwitaddress(privkey), address)
            self.assertEqual(self._coin.privtosegwitaddress(wif), address)
            self.assertEqual(self._coin.privtoaddr(wif), address)
            self.assertEqual(self._coin.pub_to_segwit_address(privtopub(privkey)), address)
            self.assertEqual(self._coin.pub_to_segwit_address(privtopub(wif)), address)
            self.assertTrue(self._coin.is_native_segwit(address))
            self.assertTrue(self._coin.is_address(address))
            self.assertFalse(self._coin.is_p2sh(address))
            self.assertFalse(self._coin.is_p2pkh(address))

    async def assertBalanceOK(self):
        result = await self._coin.get_balance(self.unspent_addresses[0])
        self.assertEqual(self.balance, result)

    async def assertGeneratorEqual(self, expected: List[Any], agen: AsyncGenerator[Any, None], order_by: str = None):
        result = await alist(agen)
        if order_by:
            expected = sorted(expected, key=lambda d: d[order_by])
            result = sorted(result, key=lambda d: d[order_by])
        self.assertEqual(expected, result)

    async def assertBalancesOK(self):
        agen = self._coin.get_balances(*self.unspent_addresses)
        await self.assertGeneratorEqual(self.balances, agen)

    async def assertBalanceMerkleProvenOK(self):
        result = await self._coin.balance_merkle_proven(self.unspent_addresses[0])
        self.assertEqual(self.balance['confirmed'], result)

    async def assertBalancesMerkleProvenOK(self):
        balances = [{'address': tx['address'], 'balance': tx['confirmed']} for tx in self.balances]
        agen = self._coin.balances_merkle_proven(*self.unspent_addresses)
        await self.assertGeneratorEqual(balances, agen)

    async def assertHistoryOK(self):
        result = await self._coin.history(self.unspent_addresses[0])
        self.assertEqual(self.history, result)

    async def assertHistoriesOK(self):
        agen = self._coin.get_histories(*self.unspent_addresses, merkle_proof=True)
        await self.assertGeneratorEqual(self.histories, agen, 'tx_hash')

    async def assertUnspentOK(self):
        result = await self._coin.unspent(self.unspent_addresses[0])
        self.assertEqual(self.unspent, result)

    async def assertUnspentsOK(self):
        unspent_outputs = self._coin.get_unspents(*self.unspent_addresses, merkle_proof=True)
        await self.assertGeneratorEqual(self.unspents, unspent_outputs, 'tx_hash')

    async def assertMixedSegwitTransactionOK(self, expected_tx_id: str = None):

        # Find which of the three addresses currently has the most coins and choose that as the sender
        segwit_max_value = 0
        segwit_sender = self.segwit_addresses[0]
        segwit_from_addr_i = 0
        segwit_unspents = []

        for i, addr in enumerate(self.segwit_addresses):
            addr_unspents = await self._coin.unspent(addr)
            value = sum(o['value'] for o in addr_unspents)
            if value > segwit_max_value:
                segwit_max_value = value
                segwit_sender = addr
                segwit_from_addr_i = i
                segwit_unspents = addr_unspents

        regular_max_value = 0
        regular_sender = None
        regular_from_addr_i = 0
        regular_unspents = []

        for i, addr in enumerate(self.addresses):
            addr_unspents = await self._coin.unspent(addr)
            value = sum(o['value'] for o in addr_unspents)
            if value > regular_max_value:
                regular_max_value = value
                regular_sender = addr
                regular_from_addr_i = i
                regular_unspents = addr_unspents

        self.assertIn(regular_sender, self.addresses)
        self.assertIn(segwit_sender, self.segwit_addresses)
        unspents = segwit_unspents + regular_unspents
        total_value = segwit_max_value + regular_max_value
        # Arbitrarily set send value, change value, receiver and change address
        send_value = int(total_value * 0.5)

        if segwit_sender == self.segwit_addresses[0]:
            receiver = self.segwit_addresses[1]
        elif segwit_sender == self.segwit_addresses[1]:
            receiver = self.segwit_addresses[2]
        else:
            receiver = self.segwit_addresses[0]

        if regular_sender == self.addresses[0]:
            change_address = self.addresses[1]
        elif regular_sender == self.addresses[1]:
            change_address = self.addresses[2]
        else:
            change_address = self.addresses[0]

        outs = [{'value': send_value, 'address': receiver}]

        # Create the transaction using all available unspents as inputs
        fee = self.fee if expected_tx_id else None
        tx = await self._coin.mktx_with_change(unspents, outs, change_addr=change_address, fee=fee)

        segwit_privkey = self.privkeys[segwit_from_addr_i]
        regular_privkey = self.privkeys[regular_from_addr_i]

        # Verify that the private key belongs to the sender address for this network
        self.assertEqual(segwit_sender, self._coin.privtop2wpkh_p2sh(segwit_privkey),
                         msg=f"Private key does not belong to address {segwit_sender} on {self._coin.display_name}")
        self.assertEqual(regular_sender, self._coin.privtoaddr(regular_privkey),
                         msg=f"Private key does not belong to address {regular_sender} on {self._coin.display_name}")

        self.assertTrue(self._coin.is_segwit_or_p2sh(segwit_sender))
        self.assertFalse(self._coin.is_segwit_or_p2sh(regular_sender))

        # Sign each input with the given private keys
        # Try signing one at a time and also with signall and make sure they give the same result

        tx2 = deepcopy(tx)

        for i in range(0, len(segwit_unspents)):
            tx = self._coin.sign(tx, i, segwit_privkey)
        for i in range(len(segwit_unspents), len(unspents)):
            tx = self._coin.sign(tx, i, regular_privkey)

        tx2 = self._coin.signall(tx2, {segwit_sender: segwit_privkey, regular_sender: regular_privkey})

        self.assertDictEqual(dict(tx), dict(tx2))
        self.assertEqual(serialize(tx), serialize(tx2))

        self.assertEqual(tx['locktime'], 0)
        self.assertEqual(tx['version'], 1)
        fee = await self._coin.calculate_fee(tx)
        self.assertGreater(fee, 0)
        self.assertLessEqual(fee, self.max_fee)
        self.assertEqual(len(tx['ins']), len(unspents))
        self.assertEqual(len(tx['outs']), 2)
        self.assertEqual(len(tx['witness']), len(unspents))
        self.assertEqual(tx['marker'], 0)
        self.assertEqual(tx['flag'], 1)

        prev_script = None

        for inp in tx['ins']:
            script = inp['script']
            self.assertIsInstance(script, str)
            self.assertIsNot(script, '')

        for o in tx['outs']:
            script = o['script']
            if prev_script:
                self.assertNotEqual(script, prev_script)
            self.assertIsInstance(script, str)
            self.assertNotEqual(script, '')
            prev_script = script

        prev_script_code = None

        for w in tx['witness']:
            script_code = w['scriptCode']
            self.assertIsInstance(script_code, str)
            if script_code:
                self.assertEqual(w['number'], 2)
                if prev_script_code:
                    self.assertNotEqual(script_code, prev_script_code)
            else:
                self.assertEqual(w['number'], 0)
            prev_script_code = script_code

        tx = serialize(tx)
        # Push the transaction to the network
        result = await self._coin.pushtx(tx)

        if expected_tx_id:
            self.assertEqual(result, expected_tx_id)    # mainnet
        else:
            self.assertTXResultOK(tx, result)           # testnet
            await asyncio.wait([
                asyncio.create_task(self._coin.wait_unspents_changed(regular_sender, unspents)),
                asyncio.create_task(self._coin.wait_unspents_changed(segwit_sender, unspents))
                ], timeout=3.5)

    async def assertSegwitTransactionOK(self, expected_tx_id: str = None):

        # Find which of the three addresses currently has the most coins and choose that as the sender
        max_value = 0
        sender = self.segwit_addresses[0]
        from_addr_i = 0
        unspents = []

        for i, addr in enumerate(self.segwit_addresses):
            addr_unspents = await self._coin.unspent(addr)
            value = sum(o['value'] for o in addr_unspents)
            if value > max_value:
                max_value = value
                sender = addr
                from_addr_i = i
                unspents = addr_unspents

        # Arbitrarily set send value, receiver and change address
        send_value = int(max_value * 0.1)

        if sender == self.segwit_addresses[0]:
            receiver = self.segwit_addresses[1]
            change_address = self.segwit_addresses[2]
        elif sender == self.segwit_addresses[1]:
            receiver = self.segwit_addresses[2]
            change_address = self.segwit_addresses[0]
        else:
            receiver = self.segwit_addresses[0]
            change_address = self.segwit_addresses[1]

        outs = [{'value': send_value, 'address': receiver}]

        # Create the transaction using all available unspents as inputs
        fee = self.fee if expected_tx_id else None
        tx = await self._coin.mktx_with_change(unspents, outs, change_addr=change_address, fee=fee)

        privkey = self.privkeys[from_addr_i]

        # Verify that the private key belongs to the sender address for this network
        self.assertEqual(sender, self._coin.privtop2wpkh_p2sh(privkey),
                         msg=f"Private key does not belong to address {sender} on {self._coin.display_name}")

        # Sign each input with the given private key
        tx = self._coin.signall(tx, privkey)

        self.assertEqual(tx['locktime'], 0)
        self.assertEqual(tx['version'], 1)
        fee = await self._coin.calculate_fee(tx)
        self.assertGreaterEqual(fee, self._coin.minimum_fee)
        self.assertEqual(len(tx['ins']), len(unspents))
        self.assertEqual(len(tx['outs']), 2)
        self.assertEqual(len(tx['witness']), len(unspents))
        self.assertEqual(tx['marker'], 0)
        self.assertEqual(tx['flag'], 1)

        prev_script = None

        for inp in tx['ins']:
            script = inp['script']
            if prev_script:
                self.assertEqual(script, prev_script)
            self.assertIsInstance(script, str)
            self.assertIsNot(script, '')
            prev_script = script

        for o in tx['outs']:
            script = o['script']
            if prev_script:
                self.assertNotEqual(script, prev_script)
            self.assertIsInstance(script, str)
            self.assertEqual(len(script), 46)
            prev_script = script

        prev_script_code = None
        for w in tx['witness']:
            self.assertEqual(w['number'], 2)
            script_code = w['scriptCode']
            if prev_script_code:
                self.assertNotEqual(script_code, prev_script_code)
            self.assertIsInstance(script_code, str)
            self.assertNotEqual(script_code, '')
            prev_script_code = script_code

        # Push the transaction to the network

        deserialize(serialize(tx))
        result = await self._coin.pushtx(tx)

        if expected_tx_id:
            self.assertEqual(result, expected_tx_id)    # mainnet
        else:
            self.assertTXResultOK(tx, result)
            await asyncio.wait_for(self._coin.wait_unspents_changed(sender, unspents), 3.5)

    async def assertNativeSegwitTransactionOK(self, expected_tx_id: str = None):

        # Find which of the three addresses currently has the most coins and choose that as the sender
        max_value = 0
        sender = self.native_segwit_addresses[0]
        from_addr_i = 0
        unspents = []

        for i, addr in enumerate(self.native_segwit_addresses):
            addr_unspents = await self._coin.unspent(addr)
            value = sum(o['value'] for o in addr_unspents)
            if value > max_value:
                max_value = value
                sender = addr
                from_addr_i = i
                unspents = addr_unspents

        # Arbitrarily set send value, change value, receiver and change address
        send_value = int(max_value * 0.1)

        if sender == self.native_segwit_addresses[0]:
            receiver = self.native_segwit_addresses[1]
            change_address = self.native_segwit_addresses[2]
        elif sender == self.native_segwit_addresses[1]:
            receiver = self.native_segwit_addresses[2]
            change_address = self.native_segwit_addresses[0]
        else:
            receiver = self.native_segwit_addresses[0]
            change_address = self.native_segwit_addresses[1]

        outs = [{'value': send_value, 'address': receiver}]

        # Create the transaction using all available unspents as inputs
        unspents = select(unspents, outs[0]['value'] + self.max_fee)
        fee = self.fee if expected_tx_id else None
        tx = await self._coin.mktx_with_change(unspents, outs, change_addr=change_address, fee=fee)

        privkey = self.privkeys[from_addr_i]

        # Verify that the private key belongs to the sender address for this network
        self.assertEqual(sender, self._coin.privtosegwitaddress(privkey),
                         msg=f"Private key does not belong to address {sender} on {self._coin.display_name}")

        # Sign each input with the given private key
        tx = self._coin.signall(tx, privkey)

        self.assertEqual(tx['locktime'], 0)
        self.assertEqual(tx['version'], 1)
        fee = await self._coin.calculate_fee(tx)
        self.assertGreater(fee, 0)
        self.assertGreaterEqual(fee, self._coin.minimum_fee)
        self.assertEqual(len(tx['ins']), len(unspents))
        self.assertEqual(len(tx['outs']), 2)
        self.assertEqual(len(tx['witness']), len(unspents))
        self.assertEqual(tx['marker'], 0)
        self.assertEqual(tx['flag'], 1)

        prev_script = None

        for inp in tx['ins']:
            script = inp['script']
            self.assertEqual(script, '')

        for o in tx['outs']:
            script = o['script']
            if prev_script:
                self.assertNotEqual(script, prev_script)
            self.assertIsInstance(script, str)
            self.assertGreaterEqual(len(script), 44)
            prev_script = script

        prev_script_code = None
        for w in tx['witness']:
            self.assertEqual(w['number'], 2)
            script_code = w['scriptCode']
            if prev_script_code:
                self.assertNotEqual(script_code, prev_script_code)
            self.assertIsInstance(script_code, str)
            self.assertNotEqual(script_code, '')
            prev_script_code = script_code

        tx = serialize(tx)

        # Push the transaction to the network
        result = await self._coin.pushtx(tx)
        if expected_tx_id:
            self.assertEqual(result, expected_tx_id)    # mainnet
        else:
            self.assertTXResultOK(tx, result)           # testnet
            await asyncio.wait_for(self._coin.wait_unspents_changed(sender, unspents), 3.5)

    async def assertCashAddressTransactionOK(self, expected_tx_id: str = None):

        # Find which of the three addresses currently has the most coins and choose that as the sender
        max_value = 0
        sender = self.cash_addresses[0]
        from_addr_i = 0
        unspents = []

        for i, addr in enumerate(self.cash_addresses):
            addr_unspents = await self._coin.unspent(addr)
            value = sum(o['value'] for o in addr_unspents)
            if value > max_value:
                max_value = value
                sender = addr
                from_addr_i = i
                unspents = addr_unspents

        # Arbitrarily set send value, change value, receiver and change address
        send_value = int(max_value * 0.1)

        if sender == self.cash_addresses[0]:
            receiver = self.cash_addresses[1]
            change_address = self.cash_addresses[2]
        elif sender == self.cash_addresses[1]:
            receiver = self.cash_addresses[2]
            change_address = self.cash_addresses[0]
        else:
            receiver = self.cash_addresses[0]
            change_address = self.cash_addresses[1]

        outs = [{'value': send_value, 'address': receiver}]

        # Create the transaction using all available unspents as inputs
        unspents = select(unspents, outs[0]['value'] + self.max_fee)
        fee = self.fee if expected_tx_id else None
        tx = await self._coin.mktx_with_change(unspents, outs, change_addr=change_address, fee=fee)

        privkey = self.privkeys[from_addr_i]

        # Verify that the private key belongs to the sender address for this network
        self.assertEqual(sender, self._coin.privtocashaddress(privkey),
                         msg=f"Private key does not belong to address {sender} on {self._coin.display_name}")

        # Sign each input with the given private key
        tx = self._coin.signall(tx, privkey)

        self.assertEqual(tx['locktime'], 0)
        self.assertEqual(tx['version'], 1)
        fee = await self._coin.calculate_fee(tx)
        self.assertGreater(fee, 0)
        self.assertLessEqual(fee, self.max_fee)
        self.assertEqual(len(tx['ins']), len(unspents))
        self.assertEqual(len(tx['outs']), 2)
        self.assertNotIn('witness', tx)
        self.assertNotIn('marker', tx)
        self.assertNotIn('flag', tx)

        prev_script = None

        for inp in tx['ins']:
            script = inp['script']
            if prev_script:
                self.assertNotEqual(script, prev_script)
            self.assertIsInstance(script, str)
            self.assertIsNot(script, '')
            prev_script = script

        for o in tx['outs']:
            script = o['script']
            if prev_script:
                self.assertNotEqual(script, prev_script)
            self.assertIsInstance(script, str)
            self.assertEqual(len(script), 50)
            prev_script = script

        # Serialize and push the transaction to the network
        tx_serialized = serialize(tx)
        result = await self._coin.pushtx(tx)
        if expected_tx_id:
            self.assertEqual(result, expected_tx_id)  # mainnet
        else:
            self.assertTXResultOK(tx_serialized, result)  # testnet
            await asyncio.wait_for(self._coin.wait_unspents_changed(sender, unspents), 3.5)


    def assertTXResultOK(self, tx: Union[str, Tx], result):
        if not isinstance(tx, str):
            tx = serialize(tx)
        tx_hash = public_txhash(tx)
        self.assertEqual(result, tx_hash)
        print("TX %s broadcasted successfully" % result)

    async def assertTransactionOK(self, expected_tx_id: str = None):

        # Find which of the three addresses currently has the most coins and choose that as the sender
        max_value = 0
        sender = self.addresses[0]
        from_addr_i = 0
        unspents = []

        for i, addr in enumerate(self.addresses):
            addr_unspents = await self._coin.unspent(addr)
            value = sum(o['value'] for o in addr_unspents)
            if value > max_value:
                max_value = value
                sender = addr
                from_addr_i = i
                unspents = addr_unspents

        # Arbitrarily set send value, receiver and change address
        send_value = int(max_value * 0.1)

        if sender == self.addresses[0]:
            receiver = self.addresses[1]
            change_address = self.addresses[2]
        elif sender == self.addresses[1]:
            receiver = self.addresses[2]
            change_address = self.addresses[0]
        else:
            receiver = self.addresses[0]
            change_address = self.addresses[1]

        outs = [{'value': send_value, 'address': receiver}]

        # Create the transaction using all available unspents as inputs

        fee = self.fee if expected_tx_id else None
        tx = await self._coin.mktx_with_change(unspents, outs, change_addr=change_address, fee=fee)

        privkey = self.privkeys[from_addr_i]

        # Verify that the private key belongs to the sender address for this network
        self.assertEqual(sender, self._coin.privtoaddr(privkey),
                         msg=f"Private key does not belong to address {sender} on {self._coin.display_name}")

        # Sign each input with the given private key
        tx = self._coin.signall(tx, privkey)

        self.assertEqual(tx['locktime'], 0)
        self.assertEqual(tx['version'], 1)
        fee = await self._coin.calculate_fee(tx)
        self.assertGreater(fee, 0)
        self.assertLessEqual(fee, self.max_fee)
        self.assertEqual(len(tx['ins']), len(unspents))
        self.assertEqual(len(tx['outs']), 2)
        self.assertNotIn('witness', tx)
        self.assertNotIn('marker', tx)
        self.assertNotIn('flag', tx)

        prev_script = None

        for inp in tx['ins']:
            script = inp['script']
            if prev_script:
                self.assertNotEqual(script, prev_script)
            self.assertIsInstance(script, str)
            self.assertIsNot(script, '')
            prev_script = script

        for o in tx['outs']:
            script = o['script']
            if prev_script:
                self.assertNotEqual(script, prev_script)
            self.assertIsInstance(script, str)
            self.assertEqual(len(script), 50)
            prev_script = script

        # Serialize and push the transaction to the network
        tx_serialized = serialize(tx)
        result = await self._coin.pushtx(tx)
        if expected_tx_id:
            self.assertEqual(result, expected_tx_id)        # mainnet
        else:
            self.assertTXResultOK(tx_serialized, result)       # testnet
            await asyncio.wait_for(self._coin.wait_unspents_changed(sender, unspents), 0.5)

    async def assertTransactionToPKOK(self, expected_tx_id: str = None):

        # Find which of the three addresses currently has the most coins and choose that as the sender
        max_value = 0
        pub_keys = [privtopub(priv) for priv in self.privkeys]
        sender = pub_keys[0]
        from_addr_i = 0
        unspents = []

        for i, pub in enumerate(pub_keys):
            addr_unspents = await self._coin.unspent(pub)
            value = sum(o['value'] for o in addr_unspents)
            if value > max_value:
                max_value = value
                sender = pub
                from_addr_i = i
                unspents = addr_unspents

        # Arbitrarily set send value, receiver and change address
        send_value = int(max_value * 0.1)

        if sender == pub_keys[0]:
            receiver = pub_keys[1]
            change_address = pub_keys[2]
        elif sender == pub_keys[1]:
            receiver = pub_keys[2]
            change_address = pub_keys[0]
        else:
            receiver = pub_keys[0]
            change_address = pub_keys[1]

        outs = [{'value': send_value, 'address': receiver}]

        # Create the transaction using all available unspents as inputs
        fee = self.fee if expected_tx_id else None
        tx = await self._coin.mktx_with_change(unspents, outs, change_addr=change_address, fee=fee)

        privkey = self.privkeys[from_addr_i]

        # Sign each input with the given private key
        tx = self._coin.signall(tx, privkey)

        self.assertEqual(tx['locktime'], 0)
        self.assertEqual(tx['version'], 1)
        fee = await self._coin.calculate_fee(tx)
        self.assertGreater(fee, 0)
        self.assertLessEqual(fee, self.max_fee)
        self.assertEqual(len(tx['ins']), len(unspents))
        self.assertEqual(len(tx['outs']), 2)
        self.assertNotIn('witness', tx)
        self.assertNotIn('marker', tx)
        self.assertNotIn('flag', tx)

        prev_script = None

        for inp in tx['ins']:
            script = inp['script']
            if prev_script:
                self.assertNotEqual(script, prev_script)
            self.assertIsInstance(script, str)
            self.assertIsNot(script, '')
            prev_script = script

        for o in tx['outs']:
            script = o['script']
            if prev_script:
                self.assertNotEqual(script, prev_script)
            self.assertIsInstance(script, str)
            self.assertTrue(any(val == len(script) for val in (70, 134)))
            prev_script = script

        # Serialize and push the transaction to the network
        tx_serialized = serialize(tx)
        result = await self._coin.pushtx(tx)
        if expected_tx_id:
            self.assertEqual(result, expected_tx_id)        # mainnet
        else:
            self.assertTXResultOK(tx_serialized, result)       # testnet
            await asyncio.wait_for(self._coin.wait_unspents_changed(sender, unspents), 0.5)

    def delete_key_by_name(self, obj, key):
        if isinstance(obj, dict):
            for k, v  in obj.items():
                if k == key:
                    del obj[k]
                    self.delete_key_by_name(obj, key)
                    break
                elif isinstance(v, (dict, list)):
                    self.delete_key_by_name(v, key)
        elif isinstance(obj, list):
            for i in obj:
                self.delete_key_by_name(i, key)

    async def assertGetTXOK(self):
        tx = await self._coin.get_tx(self.txid)
        self.assertListEqual(list(tx.keys()), ['ins', 'outs', 'version', 'locktime'])
        self.assertEqual(tx, self.tx)

    async def assertGetTxsOK(self):
        txs = await alist(self._coin.get_txs(self.txid))
        self.assertListEqual(list(txs[0].keys()),
                             ['ins', 'outs', 'version',  'locktime'])

    async def assertGetSegwitTXOK(self):
        tx = await self._coin.get_tx(self.txid)
        self.assertListEqual(list(tx.keys()), ['ins', 'outs', 'version', 'marker', 'flag', 'witness', 'locktime'])
        self.assertEqual(tx, self.tx)

    async def assertGetVerboseTXOK(self):
        tx = await self._coin.get_verbose_tx(self.txid)
        for key in ('height', 'vsize', 'weight'):
            if key in tx:
                self.expected_tx_verbose_keys.append(key)
        self.assertListEqual(sorted(tx.keys()), sorted(self.expected_tx_verbose_keys))

    async def assertTxsOK(self):
        txs = await alist(self._coin.get_txs(self.txid))
        self.assertListEqual(list(txs[0].keys()),
                             ['ins', 'outs', 'version', 'locktime'])

    async def assertGetSegwitTxsOK(self):
        txs = await alist(self._coin.get_txs(self.txid))
        self.assertListEqual(list(txs[0].keys()),
                             ['ins', 'outs', 'version', 'marker', 'flag', 'witness', 'locktime'])

    async def assertMultiSigTransactionOK(self, expected_tx_id: str = None):
        pubs = [privtopub(priv) for priv in self.privkeys]
        script1, address1 = self._coin.mk_multsig_address(*pubs, num_required=2)
        self.assertEqual(address1, self.multisig_addresses[0])
        pubs2 = pubs[0:2]
        script2, address2 = self._coin.mk_multsig_address(*pubs2)
        self.assertEqual(address2, self.multisig_addresses[1])

        # Find which of the three addresses currently has the most coins and choose that as the sender
        max_value = 0
        sender = None
        unspents = []

        for i, addr in enumerate(self.multisig_addresses):
            addr_unspents = await self._coin.unspent(addr)
            value = sum(o['value'] for o in addr_unspents)
            if value > max_value:
                max_value = value
                sender = addr
                unspents = addr_unspents

        self.assertGreater(max_value, 0)

        # Arbitrarily set send value, receiver and change address
        send_value = int(max_value * 0.5)

        receiver, script = (address2, script1) if sender == address1 else (address1, script2)

        fee = self.fee

        tx = await self._coin.preparetx(sender, receiver, send_value, fee)

        for i in range(0, len(tx['ins'])):
            if sender == address1:
                sig1 = self._coin.multisign(tx, i, script, self.privkeys[0])
                sig3 = self._coin.multisign(tx, i, script, self.privkeys[2])
                tx = self._coin.apply_multisignatures(tx, i, script, sig1, sig3)
            else:
                sig1 = self._coin.multisign(tx, i, script, self.privkeys[0])
                sig2 = self._coin.multisign(tx, i, script, self.privkeys[1])
                tx = self._coin.apply_multisignatures(tx, i, script, sig1, sig2)

        self.assertEqual(tx['locktime'], 0)
        self.assertEqual(tx['version'], 1)
        fee = await self._coin.calculate_fee(tx)
        self.assertGreaterEqual(fee, self._coin.minimum_fee)
        self.assertGreaterEqual(len(tx['ins']), 1)
        self.assertEqual(len(tx['outs']), 2)
        self.assertNotIn('witness', tx)
        self.assertNotIn('marker', tx)
        self.assertNotIn('flag', tx)

        prev_script = None

        for inp in tx['ins']:
            script = inp['script']
            if prev_script:
                self.assertNotEqual(script, prev_script)
            self.assertIsInstance(script, str)
            self.assertIsNot(script, '')
            prev_script = script

        for o in tx['outs']:
            script = o['script']
            if prev_script:
                self.assertNotEqual(script, prev_script)
            self.assertIsInstance(script, str)
            self.assertNotEqual(script, '')
            prev_script = script

        # Push the transaction to the network
        result = await self._coin.pushtx(tx)
        if expected_tx_id:
            self.assertEqual(result, expected_tx_id)    # mainnet
        else:
            self.assertTXResultOK(tx, result)           # testnet
            await asyncio.wait_for(self._coin.wait_unspents_changed(sender, unspents), 4.5)

    async def assertNativeSegwitMultiSigTransactionOK(self, expected_tx_id: str = None):
        pubs = [privtopub(priv) for priv in self.privkeys]
        script1, address1 = self._coin.mk_multsig_segwit_address(*pubs, num_required=2)
        self.assertEqual(address1, self.native_segwit_multisig_addresses[0])
        pubs2 = pubs[0:2]
        script2, address2 = self._coin.mk_multsig_segwit_address(*pubs2)
        self.assertEqual(address2, self.native_segwit_multisig_addresses[1])

        # Find which of the three addresses currently has the most coins and choose that as the sender
        max_value = 0
        sender = None
        unspents = []

        for i, addr in enumerate(self.native_segwit_multisig_addresses):
            addr_unspents = await self._coin.unspent(addr)

            value = sum(o['value'] for o in addr_unspents)
            if value > max_value:
                max_value = value
                sender = addr
                unspents = addr_unspents

        self.assertGreater(max_value, 0)
        # Arbitrarily set send value, receiver and change address
        send_value = int(max_value * 0.1)

        receiver, script = (address2, script1) if sender == address1 else (address1, script2)

        fee = self.fee if expected_tx_id else None

        tx = await self._coin.preparetx(sender, receiver, send_value, fee=fee)

        for i in range(0, len(tx['ins'])):
            if sender == address1:
                sig1 = self._coin.multisign(tx, i, script, self.privkeys[0])
                sig3 = self._coin.multisign(tx, i, script, self.privkeys[2])
                tx = self._coin.apply_multisignatures(tx, i, script, sig1, sig3)
            else:
                sig1 = self._coin.multisign(tx, i, script, self.privkeys[0])
                sig2 = self._coin.multisign(tx, i, script, self.privkeys[1])
                tx = self._coin.apply_multisignatures(tx, i, script, sig1, sig2)

        self.assertEqual(tx['locktime'], 0)
        self.assertEqual(tx['version'], 1)
        fee = await self._coin.calculate_fee(tx)
        self.assertGreaterEqual(fee, self._coin.minimum_fee)
        self.assertGreaterEqual(len(tx['ins']), 1)
        self.assertEqual(len(tx['outs']), 2)
        self.assertEqual(tx['marker'], 0)
        self.assertEqual(tx['flag'], 1)
        self.assertEqual(len(tx['witness']), len(tx['ins']))

        prev_script = None

        for inp in tx['ins']:
            script = inp['script']
            # self.assertEqual(script, '')

        for o in tx['outs']:
            script = o['script']
            if prev_script:
                self.assertNotEqual(script, prev_script)
            self.assertIsInstance(script, str)
            self.assertGreaterEqual(len(script), 44)
            prev_script = script

        prev_script_code = None
        for w in tx['witness']:
            self.assertEqual(w['number'], 4)
            script_code = w['scriptCode']
            if prev_script_code:
                self.assertNotEqual(script_code, prev_script_code)
            self.assertIsInstance(script_code, str)
            self.assertNotEqual(script_code, '')
            prev_script_code = script_code

        prev_script = None

        for inp in tx['ins']:
            script = inp['script']
            if prev_script:
                self.assertNotEqual(script, prev_script)
            self.assertIsInstance(script, str)
            # self.assertEqual(script, '')
            prev_script = script

        for o in tx['outs']:
            script = o['script']
            if prev_script:
                self.assertNotEqual(script, prev_script)
            self.assertIsInstance(script, str)
            self.assertNotEqual(script, '')
            prev_script = script

        # Push the transaction to the network
        result = await self._coin.pushtx(tx)
        if expected_tx_id:
            self.assertEqual(result, expected_tx_id)    # mainnet
        else:
            self.assertTXResultOK(tx, result)           # testnet
            await asyncio.wait_for(self._coin.wait_unspents_changed(sender, unspents), 3.5)


    async def assertCashAddressMultiSigTransactionOK(self, expected_tx_id: str = None):
        pubs = [privtopub(priv) for priv in self.privkeys]
        script1, address1 = self._coin.mk_multsig_cash_address(*pubs, num_required=2)
        self.assertEqual(address1, self.cash_multisig_addresses[0])
        pubs2 = pubs[0:2]
        script2, address2 = self._coin.mk_multsig_cash_address(*pubs2)
        self.assertEqual(address2, self.cash_multisig_addresses[1])

        # Find which of the three addresses currently has the most coins and choose that as the sender
        max_value = 0
        sender = None
        unspents = []

        for i, addr in enumerate(self.cash_multisig_addresses):
            addr_unspents = await self._coin.unspent(addr)
            value = sum(o['value'] for o in addr_unspents)
            if value > max_value:
                max_value = value
                sender = addr
                unspents = addr_unspents

        self.assertGreater(max_value, 0)

        # Arbitrarily set send value, receiver and change address
        send_value = int(max_value * 0.5)

        receiver, script = (address2, script1) if sender == address1 else (address1, script2)

        fee = self.fee

        tx = await self._coin.preparetx(sender, receiver, send_value, fee)

        for i in range(0, len(tx['ins'])):
            if sender == address1:
                sig1 = self._coin.multisign(tx, i, script, self.privkeys[0])
                sig3 = self._coin.multisign(tx, i, script, self.privkeys[2])
                tx = self._coin.apply_multisignatures(tx, i, script, sig1, sig3)
            else:
                sig1 = self._coin.multisign(tx, i, script, self.privkeys[0])
                sig2 = self._coin.multisign(tx, i, script, self.privkeys[1])
                tx = self._coin.apply_multisignatures(tx, i, script, sig1, sig2)

        self.assertEqual(tx['locktime'], 0)
        self.assertEqual(tx['version'], 1)
        fee = await self._coin.calculate_fee(tx)
        self.assertGreaterEqual(fee, self.coin.minimum_fee)
        self.assertGreaterEqual(len(tx['ins']), 1)
        self.assertEqual(len(tx['outs']), 2)
        self.assertNotIn('witness', tx)
        self.assertNotIn('marker', tx)
        self.assertNotIn('flag', tx)

        prev_script = None

        for inp in tx['ins']:
            script = inp['script']
            if prev_script:
                self.assertNotEqual(script, prev_script)
            self.assertIsInstance(script, str)
            self.assertIsNot(script, '')
            prev_script = script

        for o in tx['outs']:
            script = o['script']
            if prev_script:
                self.assertNotEqual(script, prev_script)
            self.assertIsInstance(script, str)
            self.assertNotEqual(script, '')
            prev_script = script

        # Push the transaction to the network
        result = await self._coin.pushtx(tx)
        if expected_tx_id:
            self.assertEqual(result, expected_tx_id)    # mainnet
        else:
            self.assertTXResultOK(tx, result)           # testnet
            await asyncio.wait_for(self._coin.wait_unspents_changed(sender, unspents), 2.5)

    async def assertBlockHeaderOK(self):
        blockinfo = await self._coin.block_header(self.txheight)
        block_hash = hexlify(blockinfo['hash']).decode()
        self.assertListEqual(sorted(blockinfo.keys()),
                             ['bits', 'hash', 'merkle_root', 'nonce', 'prevhash', 'timestamp', 'version']
                             )
        self.assertEqual(block_hash, self.block_hash)

    async def assertBlockHeadersOK(self):
        blockinfos = await alist(self._coin.block_headers(self.txheight))
        for blockinfo in blockinfos:
            self.assertListEqual(sorted(blockinfo.keys()),
                ['bits', 'hash', 'merkle_root', 'nonce', 'prevhash', 'timestamp', 'version']
            )
            block_hash = hexlify(blockinfo['hash']).decode()
            self.assertEqual(block_hash, self.block_hash)

    async def assertMerkleProofOK(self):
        tx = self.unspent[0]
        tx_hash = tx['tx_hash']
        proof = await self._coin.merkle_prove(self.unspent[0])
        self.assertDictEqual(dict(proof), {
            'tx_hash': tx_hash,
            'proven': True
        })

    async def assertSendMultiRecipientsTXOK(self, expected_tx_id: str = None):

        # Find which of the three addresses currently has the most coins and choose that as the sender
        max_value = 0
        sender = self.addresses[0]
        from_addr_i = 0
        unspents = []

        for i, addr in enumerate(self.addresses):
            addr_unspents = await self._coin.unspent(addr)
            value = sum(o['value'] for o in addr_unspents)
            if value > max_value:
                max_value = value
                sender = addr
                from_addr_i = i
                unspents = addr_unspents

        privkey = self.privkeys[from_addr_i]

        # Arbitrarily set send value, change value, receiver and change address
        fee = self.fee

        outputs_value = max_value - fee
        send_value1 = int(outputs_value * 0.2)
        send_value2 = int(outputs_value * 0.5)

        if sender == self.addresses[0]:
            receiver1 = self.addresses[1]
            receiver2 = self.addresses[2]
        elif sender == self.addresses[1]:
            receiver1 = self.addresses[2]
            receiver2 = self.addresses[0]
        else:
            receiver1 = self.addresses[0]
            receiver2 = self.addresses[1]

        self.assertEqual(sender, self._coin.privtoaddr(privkey),
                         msg=f"Private key does not belong to address {sender} on {self._coin.display_name}")

        outs = [{'address': receiver1, 'value': send_value1},  {'address': receiver2, 'value': send_value2}]

        result = await self._coin.send_to_multiple_receivers_tx(privkey, sender, outs, fee=fee)

        if expected_tx_id:
            self.assertEqual(result, expected_tx_id)            # mainnet
        else:
            self.assertIsInstance(result, str)                  # testnet
            print("TX %s broadcasted successfully" % result)
            await asyncio.wait_for(self._coin.wait_unspents_changed(sender, unspents), 3.5)

    async def assertSendOK(self, expected_tx_id: str = None):

        # Find which of the three addresses currently has the most coins and choose that as the sender
        max_value = 0
        sender = self.addresses[0]
        from_addr_i = 0
        unspents = []

        for i, addr in enumerate(self.addresses):
            addr_unspents = await self._coin.unspent(addr)
            value = sum(o['value'] for o in addr_unspents)
            if value > max_value:
                max_value = value
                sender = addr
                from_addr_i = i
                unspents = addr_unspents

        privkey = self.privkey_standard_wifs[from_addr_i]

        self.assertEqual(sender, self._coin.privtoaddr(privkey),
                         msg=f"Private key does not belong to address {sender} on {self._coin.display_name}")

        # Arbitrarily set send value, change value, receiver and change address
        send_value = int(max_value * 0.2)

        if sender == self.addresses[0]:
            receiver = self.addresses[1]
        elif sender == self.addresses[1]:
            receiver = self.addresses[2]
        else:
            receiver = self.addresses[0]

        result = await self._coin.send(privkey, sender, receiver, send_value)
        if expected_tx_id:
            self.assertEqual(result, expected_tx_id)    # mainnet
        else:
            self.assertIsInstance(result, str)  # testnet
            print("TX %s broadcasted successfully" % result)
            await asyncio.wait_for(self._coin.wait_unspents_changed(sender, unspents), 10)

    async def assertSubscribeBlockHeadersOK(self):
        queue = asyncio.Queue()
        block_keys = ['bits', 'hash', 'merkle_root', 'nonce', 'prevhash', 'timestamp', 'version']

        async def on_new_block(height: int, hex_header: str, header: BlockHeader) -> None:
            await queue.put((height, hex_header, header))

        await self._coin.subscribe_to_block_headers(on_new_block)
        result = await queue.get()
        height, hex_header, header = result
        self.assertGreater(height, self.min_latest_height)
        self.assertEqual(deserialize_header(binascii.unhexlify(hex_header)), header)
        self.assertListEqual(sorted(header.keys()), block_keys)
        await self._coin.unsubscribe_from_block_headers()

    async def assertSubscribeBlockHeadersSyncCallbackOK(self):
        queue = Queue()
        block_keys = ['bits', 'hash', 'merkle_root', 'nonce', 'prevhash', 'timestamp', 'version']

        def on_new_block(height: int, hex_header: str, header: BlockHeader) -> None:
            queue.put_nowait((height, hex_header, header))

        await self._coin.subscribe_to_block_headers(on_new_block)
        result = await asyncio.get_event_loop().run_in_executor(None, queue.get)
        height, hex_header, header = result
        self.assertGreater(height, self.min_latest_height)
        self.assertEqual(deserialize_header(binascii.unhexlify(hex_header)), header)
        self.assertListEqual(sorted(header.keys()), block_keys)
        await self._coin.unsubscribe_from_block_headers()

    async def assertLatestBlockOK(self):
        height, hex_header, header = await self._coin.block
        self.assertGreater(height, self.min_latest_height)
        self.assertIsInstance(hex_header, str)
        self.assertIsInstance(header, dict)
        height, hex_header, header = await self._coin.block
        self.assertGreater(height, self.min_latest_height)
        self.assertIsInstance(hex_header, str)
        self.assertIsInstance(header, dict)

    async def assertConfirmationsOK(self):
        confirmations = await self._coin.confirmations(0)
        block = await self._coin.block
        height = block[0]
        self.assertEqual(confirmations, 0)
        confirmations = await self._coin.confirmations(height - 1)
        self.assertEqual(confirmations, 2)
        confirmations = await self._coin.confirmations(1)
        self.assertEqual(confirmations, height)

    async def assertSubscribeAddressOK(self):
        queue = asyncio.Queue()
        address = self.addresses[0]

        async def add_to_queue(addr: str, status: str) -> None:
            await queue.put((addr, status))

        await self._coin.subscribe_to_address(add_to_queue, address)
        addr, initial_status = await queue.get()
        self.assertEqual(addr, address)
        await self.assertTransactionOK()
        addr, status = await queue.get()
        self.assertEqual(addr, address)
        self.assertNotEqual(initial_status, status)
        await self._coin.unsubscribe_from_address(address)

    async def assertSubscribeAddressSyncCallbackOK(self):
        queue = Queue()
        try:
            address = self.segwit_addresses[0]
        except IndexError:
            try:
                address = self.cash_addresses[0]
            except IndexError:
                address = self.addresses[0]


        def add_to_queue(addr: str, status: str) -> None:
            queue.put((addr, status))

        await self._coin.subscribe_to_address(add_to_queue, address)
        addr, initial_status = await asyncio.get_event_loop().run_in_executor(None, queue.get)
        self.assertEqual(addr, address)
        if self.segwit_addresses:
            await self.assertSegwitTransactionOK()
        elif self.cash_addresses:
            await self.assertCashAddressTransactionOK()
        else:
            await self.assertTransactionOK()
        addr, status = await asyncio.get_event_loop().run_in_executor(None, queue.get)
        self.assertEqual(addr, address)
        self.assertNotEqual(initial_status, status)
        await self._coin.unsubscribe_from_address(address)

    async def assertSubscribeAddressTransactionsOK(self):
        queue = asyncio.Queue()
        address = self.addresses[0]

        async def add_to_queue(address: str, txs: List[Tx], newly_confirmed: List[Tx], history: List[Tx],
                               unspent: List[Tx], confirmed: int, unconfirmed: int, proven: int) -> None:
            await queue.put((address, txs, newly_confirmed, history, unspent, confirmed, unconfirmed, proven))

        await self._coin.subscribe_to_address_transactions(add_to_queue, address)
        addr, start_txs, start_newly_confirmed, start_history, unspent, start_confirmed, start_unconfirmed, start_proven = await queue.get()
        self.assertEqual(addr, address)
        self.assertEqual(start_txs, [])
        self.assertEqual(start_newly_confirmed, [])
        await self.assertTransactionOK()
        addr, new_txs, newly_confirmed, history, unspent, confirmed, unconfirmed, proven = await queue.get()
        self.assertEqual(addr, address)
        self.assertGreaterEqual(len(new_txs), 1)
        self.assertEqual(len(newly_confirmed), 0)
        self.assertGreaterEqual(len(history), 1)
        self.assertEqual(len(start_history), len(history))
        self.assertNotEqual(unconfirmed, start_unconfirmed)
        await self._coin.unsubscribe_from_address(address)

    async def assertSubscribeAddressTransactionsSyncOK(self):
        queue = Queue()
        address = self.addresses[0]

        def add_to_queue(address: str, txs: List[Tx], newly_confirmed: List[Tx], history: List[Tx],
                         unspent: List[Tx], confirmed: int, unconfirmed: int, proven: int) -> None:
            queue.put((address, txs, newly_confirmed, history, unspent, confirmed, unconfirmed, proven))

        await self._coin.subscribe_to_address_transactions(add_to_queue, address)
        addr, start_txs, start_newly_confirmed, start_history, start_unspent, start_confirmed, start_unconfirmed, start_proven = await asyncio.get_event_loop().run_in_executor(None, queue.get)
        self.assertEqual(addr, address)
        self.assertEqual(start_txs, [])
        self.assertEqual(start_newly_confirmed, [])
        await self.assertTransactionOK()
        addr, new_txs, newly_confirmed, history, unspent, confirmed, unconfirmed, proven = await asyncio.get_event_loop().run_in_executor(
            None, queue.get)
        self.assertEqual(addr, address)
        self.assertGreaterEqual(len(new_txs), 1)
        self.assertEqual(len(newly_confirmed), 0)
        self.assertGreaterEqual(len(history), 1)
        self.assertEqual(len(start_history), len(history))
        self.assertNotEqual(unconfirmed, start_unconfirmed)
        await self._coin.unsubscribe_from_address(address)



from .bitcoin import BaseCoin


class Doge(BaseCoin):
    coin_symbol = "DOGE"
    display_name = "Dogecoin"
    segwit_supported = False
    magicbyte = 0x1e
    minimum_fee = 300000
    script_magicbyte = 0x16
    wif_prefix: int = 0x9e
    segwit_hrp = "doge"
    hd_path = 3
    client_kwargs = {
        'server_file': 'doge.json',
        'use_ssl': False
    }
    xpriv_prefix = 0x02facafd
    xpub_prefix = 0x02fac398
    testnet_overrides = {
        'display_name': "Dogecoin Testnet",
        'coin_symbol': "Dogecoin",
        'magicbyte': 0x71,
        'script_magicbyte': 0xc4,
        'hd_path': 1,
        'wif_prefix': 0xf1,
        'segwit_hrp': 'xdoge',
        'minimum_fee': 300000,
        'client_kwargs': {
            'server_file': 'doge_testnet.json',
            'use_ssl': False
        },
        'xpriv_prefix': 0x04358394,
        'xpub_prefix': 0x043587cf
    }



from cryptos.coins_async import BaseCoin, Bitcoin, BitcoinCash, Dash, Litecoin, Doge


coins = {c.coin_symbol: c for c in (Bitcoin, Litecoin, BitcoinCash, Dash, Doge)}


def get_coin(coin_symbol: str, testnet: bool) -> BaseCoin:
    symbol = coin_symbol.upper()
    return coins[symbol](testnet=testnet)


coin_list = [c.lower() for c in coins.keys()]



#!/usr/bin/python
import copy
from .main import *
import binascii
from copy import deepcopy
from .opcodes import opcodes
from . import segwit_addr
from . import cashaddr
from _functools import reduce
from .utils import is_hex

from typing import AnyStr, Union
from .types import Tx, Witness


### Hex to bin converter and vice versa for objects


def json_is_base(obj, base):
    if isinstance(obj, bytes):
        return False
    
    alpha = get_code_string(base)
    if isinstance(obj, string_types):
        for i in range(len(obj)):
            if alpha.find(obj[i]) == -1:
                return False
        return True
    elif isinstance(obj, int_types) or obj is None:
        return True
    elif isinstance(obj, list):
        for i in range(len(obj)):
            if not json_is_base(obj[i], base):
                return False
        return True
    else:
        for x in obj:
            if not json_is_base(obj[x], base):
                return False
        return True


def json_changebase(obj, changer):
    if isinstance(obj, string_or_bytes_types):
        return changer(obj)
    elif isinstance(obj, int_types) or obj is None:
        return obj
    elif isinstance(obj, list):
        return [json_changebase(x, changer) for x in obj]
    return dict((x, json_changebase(obj[x], changer)) for x in obj)

# Hashing transactions for signing

SIGHASH_ALL = 1
SIGHASH_NONE = 2
SIGHASH_SINGLE = 3
# this works like SIGHASH_ANYONECANPAY | SIGHASH_ALL, might as well make it explicit while
# we fix the constant
SIGHASH_ANYONECANPAY = 0x81
SIGHASH_FORKID = 0x40


def encode_1_byte(val):
    return encode(val, 256, 1)[::-1]


def encode_4_bytes(val):
    return encode(val, 256, 4)[::-1]


def encode_8_bytes(val):
    return encode(val, 256, 8)[::-1]


def list_to_bytes(vals: List[bytes]) -> bytes:
    try:
        return reduce(lambda x, y: x + y, vals, bytes())
    except Exception as e:
        print(e)
        pass


def dbl_sha256_list(vals):
    return bin_dbl_sha256(list_to_bytes(vals))


# Transaction serialization and deserialization

def is_segwit(tx: bytes) -> bool:
    """
    Checks that the marker in a transaction is set to 0. For legacy transactions this would indicate the number of
    inputs so will be at least 1.
    """
    return tx[4] == 0


def deserialize(tx: AnyStr) -> Tx:
    if isinstance(tx, str) and is_hex(tx):
        # tx = bytes(bytearray.fromhex(tx))
        return json_changebase(deserialize(binascii.unhexlify(tx)),
                               lambda x: safe_hexlify(x))
    # http://stackoverflow.com/questions/4851463/python-closure-write-to-variable-in-parent-scope
    # Python's scoping rules are demented, requiring me to make pos an object
    # so that it is call-by-reference
    pos = [0]

    def read_as_int(bytez: int):
        pos[0] += bytez
        return decode(tx[pos[0] - bytez:pos[0]][::-1], 256)

    def read_var_int() -> int:
        pos[0] += 1
        val = from_byte_to_int(tx[pos[0] - 1])
        if val < 253:
            return val
        return read_as_int(pow(2, val - 252))

    def read_bytes(bytez: int) -> str:
        pos[0] += bytez
        return tx[pos[0] - bytez:pos[0]]

    def read_var_string() -> str:
        size = read_var_int()
        return read_bytes(size)

    def read_witness_script_code() -> str:
        size = read_var_int()
        return num_to_var_int(size)+read_bytes(size)

    obj: Tx = {"ins": [], "outs": [], "version": read_as_int(4)}
    has_witness = is_segwit(tx)
    if has_witness:
        obj['marker'] = read_as_int(1)
        obj['flag'] = read_as_int(1)
    ins = read_var_int()
    for _ in range(ins):
        obj["ins"].append({
            "tx_hash": read_bytes(32)[::-1],
            "tx_pos": read_as_int(4),
            "script": read_var_string(),
            "sequence": read_as_int(4)
        })
    outs = read_var_int()
    for _ in range(outs):
        obj["outs"].append({
            "value": read_as_int(8),
            "script": read_var_string()
        })
    if has_witness:
        obj['witness'] = []
        for _ in range(ins):
            number = read_var_int()
            script_code = [read_witness_script_code() for _ in range(number)]
            obj['witness'].append({
                'number': number,
                'scriptCode': list_to_bytes(script_code)
            })
    obj["locktime"] = read_as_int(4)
    return obj


def serialize(txobj: Tx, include_witness: bool = True) -> AnyStr:
    txobj = deepcopy(txobj)
    for i in txobj['ins']:
        if 'address' in i:
            del i['address']
    if isinstance(txobj, bytes):
        txobj = bytes_to_hex_string(txobj)
    o = []
    if json_is_base(txobj, 16):
        json_changedbase = json_changebase(txobj, lambda x: binascii.unhexlify(x))
        hexlified = safe_hexlify(serialize(json_changedbase, include_witness=include_witness))
        return hexlified
    o.append(encode_4_bytes(txobj["version"]))
    if include_witness and all(k in txobj.keys() for k in ['marker', 'flag']):
        o.append(encode_1_byte(txobj["marker"]))
        o.append(encode_1_byte(txobj["flag"]))
    o.append(num_to_var_int(len(txobj["ins"])))
    for inp in txobj["ins"]:
        o.append(inp["tx_hash"][::-1])
        o.append(encode_4_bytes(inp["tx_pos"]))
        o.append(num_to_var_int(len(inp["script"])) + (inp["script"] if inp["script"] else bytes()))
        o.append(encode_4_bytes(inp["sequence"]))
    o.append(num_to_var_int(len(txobj["outs"])))
    for out in txobj["outs"]:
        o.append(encode_8_bytes(out["value"]))
        o.append(num_to_var_int(len(out["script"])) + out["script"])
    if include_witness and "witness" in txobj.keys():
        for witness in txobj["witness"]:
            o.append(num_to_var_int(witness["number"]) + (witness["scriptCode"] if witness["scriptCode"] else bytes()))
    o.append(encode_4_bytes(txobj["locktime"]))
    return list_to_bytes(o)


def uahf_digest(txobj: Tx, i: int) -> bytes:
    for inp in txobj['ins']:
        inp.pop('address', None)
    if isinstance(txobj, bytes):
        txobj = bytes_to_hex_string(txobj)
    o = []

    if json_is_base(txobj, 16):
        txobj = json_changebase(txobj, lambda x: binascii.unhexlify(x))
    o.append(encode(txobj["version"], 256, 4)[::-1])

    serialized_ins = []
    for inp in txobj["ins"]:
        serialized_ins.append(inp["tx_hash"][::-1])
        serialized_ins.append(encode_4_bytes(inp["tx_pos"]))
    inputs_hashed = dbl_sha256_list(serialized_ins)
    o.append(inputs_hashed)

    sequences = dbl_sha256_list([encode_4_bytes(inp["sequence"]) for inp in txobj['ins']])
    o.append(sequences)

    inp = txobj['ins'][i]
    o.append(inp["tx_hash"][::-1])
    o.append(encode_4_bytes(inp["tx_pos"]))
    o.append(num_to_var_int(len(inp["script"])) + (inp["script"] if inp["script"] else bytes()))
    o.append(encode_8_bytes(inp['value']))
    o.append(encode_4_bytes(inp['sequence']))

    serialized_outs = []
    for out in txobj["outs"]:
        serialized_outs.append(encode_8_bytes(out["value"]))
        serialized_outs.append(num_to_var_int(len(out["script"])) + out["script"])
    outputs_hashed = dbl_sha256_list(serialized_outs)
    o.append(outputs_hashed)

    o.append(encode_4_bytes(txobj["locktime"]))
    # o.append(b'\x01\x00\x00\x00')
    return list_to_bytes(o)


def signature_form(tx: Union[AnyStr, Tx], i: int, script, hashcode: int = SIGHASH_ALL, segwit: bool = False) -> bytes:
    i, hashcode = int(i), int(hashcode)
    if isinstance(tx, string_or_bytes_types):
        tx = deserialize(tx)
    newtx = deepcopy(tx)
    for j, inp in enumerate(newtx["ins"]):
        if j == i:
            newtx['ins'][j]['script'] = script
        else:
            newtx['ins'][j]['script'] = ""
    if segwit or hashcode & 255 == SIGHASH_ALL + SIGHASH_FORKID:
        return uahf_digest(newtx, i)
    elif hashcode == SIGHASH_NONE:
        newtx["outs"] = []
    elif hashcode == SIGHASH_SINGLE:
        newtx["outs"] = newtx["outs"][:len(newtx["ins"])]
        for out in newtx["outs"][:len(newtx["ins"]) - 1]:
            out['value'] = 2**64 - 1
            out['script'] = ""
    elif hashcode == SIGHASH_ANYONECANPAY:
        newtx["ins"] = [newtx["ins"][i]]

    return serialize(newtx, include_witness=False)

# Making the actual signatures


def der_encode_sig(v, r, s):
    b1, b2 = safe_hexlify(encode(r, 256)), safe_hexlify(encode(s, 256))
    if len(b1) and b1[0] in '89abcdef':
        b1 = '00' + b1
    if len(b2) and b2[0] in '89abcdef':
        b2 = '00' + b2
    left = '02'+encode(len(b1)//2, 16, 2)+b1
    right = '02'+encode(len(b2)//2, 16, 2)+b2
    return '30'+encode(len(left+right)//2, 16, 2)+left+right


def der_decode_sig(sig):
    leftlen = decode(sig[6:8], 16)*2
    left = sig[8:8+leftlen]
    rightlen = decode(sig[10+leftlen:12+leftlen], 16)*2
    right = sig[12+leftlen:12+leftlen+rightlen]
    return None, decode(left, 16), decode(right, 16)


def is_bip66(sig: str) -> bool:
    """Checks hex DER sig for BIP66 consistency"""
    #https://raw.githubusercontent.com/bitcoin/bips/master/bip-0066.mediawiki
    #0x30  [total-len]  0x02  [R-len]  [R]  0x02  [S-len]  [S]  [sighash]
    sig = bytearray.fromhex(sig) if is_hex(sig) else bytearray(sig)
    if (sig[0] == 0x30) and (sig[1] == len(sig)-2):     # check if sighash is missing
            sig.extend(b"\1")		                   	# add SIGHASH_ALL for testing
    #assert (sig[-1] & 124 == 0) and (not not sig[-1]), "Bad SIGHASH value"
    
    if len(sig) < 9 or len(sig) > 73: return False
    if (sig[0] != 0x30): return False
    if (sig[1] != len(sig)-3): return False
    rlen = sig[3]
    if (5+rlen >= len(sig)): return False
    slen = sig[5+rlen]
    if (rlen + slen + 7 != len(sig)): return False
    if (sig[2] != 0x02): return False
    if (rlen == 0): return False
    if (sig[4] & 0x80): return False
    if (rlen > 1 and (sig[4] == 0x00) and not (sig[5] & 0x80)): return False
    if (sig[4+rlen] != 0x02): return False
    if (slen == 0): return False
    if (sig[rlen+6] & 0x80): return False
    if (slen > 1 and (sig[6+rlen] == 0x00) and not (sig[7+rlen] & 0x80)):
        return False
    return True


def txhash(tx: AnyStr, hashcode: int = None, wtxid: bool = True) -> str:
    if isinstance(tx, str) and re.match('^[0-9a-fA-F]*$', tx):
        tx = changebase(tx, 16, 256)
    if isinstance(tx, string_or_bytes_types):
        segwit = is_segwit(tx)
    else:
        segwit = False
    if not wtxid and segwit:
        tx = serialize(deserialize(tx), include_witness=False)
    if hashcode:
        return dbl_sha256(from_string_to_bytes(tx) + encode(int(hashcode), 256, 4)[::-1])
    else:
        return safe_hexlify(bin_dbl_sha256(tx)[::-1])


def public_txhash(tx: AnyStr, hashcode: int = None) -> str:
    return txhash(tx, hashcode=hashcode, wtxid=False)


def bin_txhash(tx: AnyStr, hashcode: int = None) -> bytes:
    return binascii.unhexlify(txhash(tx, hashcode))


def ecdsa_tx_sign(tx, priv, hashcode=SIGHASH_ALL):
    rawsig = ecdsa_raw_sign(bin_txhash(tx, hashcode), priv)
    return der_encode_sig(*rawsig)+encode(hashcode & 255, 16, 2)


def ecdsa_tx_verify(tx, sig, pub, hashcode=SIGHASH_ALL):
    return ecdsa_raw_verify(bin_txhash(tx, hashcode), der_decode_sig(sig), pub)


def ecdsa_tx_recover(tx, sig, hashcode=SIGHASH_ALL):
    z = bin_txhash(tx, hashcode)
    _, r, s = der_decode_sig(sig)
    left = ecdsa_raw_recover(z, (0, r, s))
    right = ecdsa_raw_recover(z, (1, r, s))
    return (encode_pubkey(left, 'hex'), encode_pubkey(right, 'hex'))

# Scripts


def mk_pubkey_script(pubkey_hash: str) -> str:
    """
    Used in converting public key hash to input or output script
    """
    return opcodes.OP_DUP.hex() + opcodes.OP_HASH160.hex() + '14' + pubkey_hash + opcodes.OP_EQUALVERIFY.hex() + opcodes.OP_CHECKSIG.hex()


def addr_to_pubkey_script(addr: str) -> str:
    """
    Used in converting public key hash address to input or output script
    """
    magicbyte, bin = b58check_to_hex(addr)
    return mk_pubkey_script(bin)


def mk_p2pk_script(pub: str) -> str:
    """
    Used in converting public key to p2pk script
    """
    length = hex(int(len(pub) / 2)).split('0x')[1]
    return length + pub + opcodes.OP_CHECKSIG.hex()


def script_to_pk(script: str) -> str:
    """
    Used in converting p2pk script to public key
    """
    length = int(script[0:2], 16)
    return script[2: (length + 1) * 2]


def hash_to_scripthash_script(hashbin: str) -> str:
    return opcodes.OP_HASH160.hex() + '14' + hashbin + opcodes.OP_EQUAL.hex()


def mk_scripthash_script(addr: str):
    """
    Used in converting p2sh address to output script
    """
    magicbyte, hashbin = b58check_to_hex(addr)
    return hash_to_scripthash_script(hashbin)


def output_script_to_address(script, magicbyte: int = 0, script_magicbyte: int = 5,
                             segwit_hrp: str = None, cash_hrp: str = None ) -> AnyStr:
    if script.startswith('76a914') and script.endswith('88ac'):
        script = script[6:][:-4]
        return bin_to_b58check(safe_from_hex(script), magicbyte=magicbyte)
    elif script.startswith('a914') and script.endswith('87'):
        script = script[4:][:-2]
        return bin_to_b58check(safe_from_hex(script), magicbyte=script_magicbyte)
    elif script.startswith('0') and segwit_hrp:
        return decode_p2w_scripthash_script(script, 0, segwit_hrp)
    elif script.startswith('0') and cash_hrp:
        return decode_cash_scripthash_script(script, 0, cash_hrp)
    elif script.startswith('6a'):
        return binascii.unhexlify("Arbitrary Data: %s" % script[2:].decode('utf-8', 'ignore'))
    raise Exception('Unable to convert script to an address: %s' % script)


def decode_p2w_scripthash_script(script, witver, segwit_hrp):
    witprog = safe_from_hex(script[4:])
    return segwit_addr.encode_segwit_address(segwit_hrp, witver, witprog)


def decode_cash_scripthash_script(script, witver, hrp):
    witprog = safe_from_hex(script[4:])
    return cashaddr.encode(hrp, witver, witprog)


def mk_p2w_scripthash_script(witver: int, witprog: List[int]) -> str:
    """
    Used in converting a decoded pay to witness script hash address to output script
    """
    assert (0 <= witver <= 16)
    OP_n = witver + int(opcodes.OP_RESERVED) if witver > 0 else 0
    length = len(witprog)
    len_hex = hex(length).split('0x')[1]
    return bytes_to_hex_string([OP_n]) + len_hex + (bytes_to_hex_string(witprog))


def mk_p2wpkh_redeemscript(pubkey: str) -> str:
    """
    Used in converting public key to p2wpkh script
    """
    return '16' + opcodes.OP_0.hex() + '14' + pubkey_to_hash_hex(pubkey)


def mk_p2wpkh_script(pubkey: str) -> str:
    """
    Used in converting public key to p2wpkh script
    """
    script = mk_p2wpkh_redeemscript(pubkey)[2:]
    return opcodes.OP_HASH160.hex() + '14' + hex_to_hash160(script) + opcodes.OP_EQUAL.hex()


def mk_p2wpkh_scriptcode(pubkey):
    """
    Used in signing for tx inputs
    """
    return opcodes.OP_DUP.hex() + opcodes.OP_HASH160.hex() + '14' + pubkey_to_hash_hex(
        pubkey) + opcodes.OP_EQUALVERIFY.hex() + opcodes.OP_CHECKSIG.hex()


def p2wpkh_nested_script(pubkey):
    return opcodes.OP_0.hex() + '14' + hash160(safe_from_hex(pubkey))

# Output script to address representation


def deserialize_script(script):
    if isinstance(script, str) and is_hex(script):
       return json_changebase(deserialize_script(binascii.unhexlify(script)),
                              lambda x: safe_hexlify(x))
    out, pos = [], 0
    while pos < len(script):
        code = from_byte_to_int(script[pos])
        if code == 0:
            out.append(None)
            pos += 1
        elif code <= 75:
            out.append(script[pos+1:pos+1+code])
            pos += 1 + code
        elif code <= 78:
            szsz = pow(2, code - 76)
            sz = decode(script[pos+szsz: pos:-1], 256)
            out.append(script[pos + 1 + szsz:pos + 1 + szsz + sz])
            pos += 1 + szsz + sz
        elif code <= 96:
            out.append(code - 80)
            pos += 1
        else:
            out.append(code)
            pos += 1
    return out


def serialize_script_unit(unit):
    if isinstance(unit, int):
        if unit < 16:
            return from_int_to_byte(unit + 80)
        else:
            return from_int_to_byte(unit)
    elif unit is None:
        return b'\x00'
    else:
        if len(unit) <= 75:
            return from_int_to_byte(len(unit))+unit
        elif len(unit) < 256:
            return from_int_to_byte(76)+from_int_to_byte(len(unit))+unit
        elif len(unit) < 65536:
            return from_int_to_byte(77)+encode(len(unit), 256, 2)[::-1]+unit
        else:
            return from_int_to_byte(78)+encode(len(unit), 256, 4)[::-1]+unit


def serialize_script(script) -> AnyStr:
    if json_is_base(script, 16):
        return safe_hexlify(serialize_script(json_changebase(script,
                            lambda x: binascii.unhexlify(x))))

    result = bytes()
    for b in map(serialize_script_unit, script):
        result += b if isinstance(b, bytes) else bytes(b, 'utf-8')
    return result


def mk_multisig_script(*args):  # [pubs],k or pub1,pub2...pub[n],M
    """
    :param args: List of public keys to used to create multisig and M, the number of signatures required to spend
    :return: multisig script
    """
    if isinstance(args[0], list):
        pubs, M = args[0], int(args[1])
    else:
        pubs = list(filter(lambda x: len(str(x)) >= 32, args))
        M = int(args[len(pubs)])
    N = len(pubs)
    return serialize_script([M]+pubs+[N]+[opcodes.OP_CHECKMULTISIG])


# Signing and verifying

def verify_tx_input(tx, i, script, sig, pub):
    if is_hex(tx):
        tx = binascii.unhexlify(tx)
    if is_hex(script):
        script = binascii.unhexlify(script)
    if isinstance(script, string_types) and is_hex(script):
        sig = safe_hexlify(sig)
    hashcode = decode(sig[-2:], 16)
    modtx = signature_form(tx, int(i), script, hashcode)
    return ecdsa_tx_verify(modtx, sig, pub, hashcode)


def multisign(tx, i: int, script, pk, hashcode: int = SIGHASH_ALL, segwit: bool = False):
    modtx = signature_form(tx, i, script, hashcode, segwit=segwit)
    return ecdsa_tx_sign(modtx, pk, hashcode)


def apply_multisignatures(txobj: Union[Tx, str], i: int, script, *args, segwit: bool = False):
    # tx,i,script,sigs OR tx,i,script,sig1,sig2...,sig[n]
    sigs = args[0] if isinstance(args[0], list) else list(args)

    if isinstance(script, str) and re.match('^[0-9a-fA-F]*$', script):
        script = binascii.unhexlify(script)
    sigs = [binascii.unhexlify(x) if x[:2] == '30' else x for x in sigs]
    if not isinstance(txobj, dict):
        txobj = deserialize(txobj)
    if isinstance(txobj, str) and re.match('^[0-9a-fA-F]*$', txobj):
        return safe_hexlify(serialize(apply_multisignatures(binascii.unhexlify(txobj), i, script, sigs)))

    if not isinstance(txobj, dict):
        txobj = deserialize(txobj)

    if segwit:
        if 'witness' not in txobj.keys():
            txobj.update({"marker": 0, "flag": 1, "witness": []})
            for _ in range(0, i):
                witness: Witness = {"number": 0, "scriptCode": ''}
                # Pycharm IDE gives a type error for the following line, no idea why...
                # noinspection PyTypeChecker
                txobj["witness"].append(witness)
        txobj["ins"][i]["script"] = ''
        number = len(sigs) + 2
        scriptSig = safe_hexlify(serialize_script([None] + sigs + [len(script)] + deserialize_script(script)))
        witness: Witness = {"number": number, "scriptCode": scriptSig}
        # Pycharm IDE gives a type error for the following line, no idea why...
        # noinspection PyTypeChecker
        txobj["witness"].append(witness)
    else:
        # Not pushing empty elements on the top of the stack if passing no
        # script (in case of bare multisig inputs there is no script)
        script_blob = [] if script.__len__() == 0 else [script]
        scriptSig = safe_hexlify(serialize_script([None] + sigs + script_blob))
        txobj["ins"][i]["script"] = scriptSig
        if "witness" in txobj.keys():
            witness: Witness = {"number": 0, "scriptCode": ''}
            # Pycharm IDE gives a type error for the following line, no idea why...
            # noinspection PyTypeChecker
            txobj["witness"].append(witness)
    return txobj


def select(unspents, value: int):
    value = int(value)
    high = [u for u in unspents if u["value"] >= value]
    high.sort(key=lambda u: u["value"])
    low = [u for u in unspents if u["value"] < value]
    low.sort(key=lambda u: -u["value"])
    if len(high):
        return [high[0]]
    i, tv = 0, 0
    while tv < value and i < len(low):
        tv += low[i]["value"]
        i += 1
    if tv < value:
        raise Exception("Not enough funds")
    return low[:i]




from cryptos import coins_async
from cryptos.testing.testcases_async import BaseAsyncCoinTestCase
from cryptos.electrumx_client.types import ElectrumXMultiBalanceResponse
from typing import List
from unittest import mock


class TestDash(BaseAsyncCoinTestCase):
    name = "Dash"
    coin = coins_async.Dash

    addresses = ["XwPJ2c8dJifpZ422ogWaM6etzm9qZ7RyBz",
                 "Xhu5S5VibUsxjkUYoJ1RDotx339EEL1qGH",
                 "XgmCkSxeLGfe9PDnemqx1SzuAS71BPYDoY"]
    multisig_addresses: List[str] = ["7VvjrtmMoXdUNXWKMfeXTdHRNXDhhs16sQ", "7TcYXDLZRJc3WS3gAHUehsh5q6Zb4KSUic"]
    privkeys = ["098ddf01ebb71ead01fc52cb4ad1f5cafffb5f2d052dd233b3cad18e255e1db1",
                "0861e1bb62504f5e9f03b59308005a6f2c12c34df108c6f7c52e5e712a08e91401",
                "c396c62dfdc529645b822dc4eaa7b9ddc97dd8424de09ca19decce61e6732f71"]  #Private keys for above address_derivations in same order
    privkey_standard_wifs: List[str] = ["7qd538kkwpDfEC13gqUju9maWGqcjCAeWhsCDzJ1s7WXQCRhBqM",
                                        "XBZvhbM7yLHg2pTGzgK9f4PDWB7AA48Rd8psZHiYEpdgDBikbLbe",
                                        "7s313e2yHvdGx45ydMuL1UHHcZeQAPQs5n62x8H76HjACPvsS4x"]
    unspent_addresses = ["XtuVUju4Baaj7YXShQu4QbLLR7X2aw9Gc8"]
    unspent = [
        {'address': 'XtuVUju4Baaj7YXShQu4QbLLR7X2aw9Gc8', 'height': 1086570,
         'tx_hash': '47266dc659f7271d26dd2b10369895c26d7b16b7db7cd577b60896c7c6cc1974', 'tx_pos': 0,
         'value': 99808}]
    unspents = unspent
    balance: ElectrumXMultiBalanceResponse = {'confirmed': 99808, 'unconfirmed': 0}
    balances: List[ElectrumXMultiBalanceResponse] =[{'address': unspent_addresses[0]} | dict(balance)]
    history = [{'height': 1086570, 'tx_hash': '47266dc659f7271d26dd2b10369895c26d7b16b7db7cd577b60896c7c6cc1974'}]
    histories = [{'address': 'XtuVUju4Baaj7YXShQu4QbLLR7X2aw9Gc8', 'height': 1086570,
                  'tx_hash': '47266dc659f7271d26dd2b10369895c26d7b16b7db7cd577b60896c7c6cc1974'}]
    fee: int = 500
    max_fee: int = 3500
    testnet = False
    min_latest_height = 801344
    txheight: int = 509045
    block_hash = "0134634c7f95ebc31f3929d30a334f7d5c87a41625b64e26bf680028fa47fc63"
    txid = "e7a607c5152863209f33cec4cc0baed973f7cfd75ae28130e623c099fde7072c"
    raw_tx = "0100000001b8f99a18e65963d07b239c77528ac0521b607e10ae3d82d2d417283fd61273ac000000006b483045022100dd5bce7f3898d754813e3b77f25fb9731a3d4e3fa7877d9313db9d728a5b28df02203be0bc1166c7a5669d0f8134b9dcc3a2c5fcf69e4a9106780f9ea3b6c9c60edd01210301fb572fa47ec4de2c8cecc1a38fe4fedc2c533dc3a9ae126fb8a225c8d895f5ffffffff02a5d93d10000000001976a9144ce676449e20eb2e329173d34b752fd70f02299088ac00ef1c0d000000001976a914561a5126d9674f51688677c7d19127bd6386795988ac00000000"
    expected_tx_verbose_keys: List[str] = ['blockhash', 'blocktime', 'chainlock', 'confirmations', 'hex', 'instantlock', 'instantlock_internal', 'locktime', 'size', 'time', 'txid', 'type', 'version', 'vin', 'vout']

    def test_standard_wif_ok(self):
        self.assertStandardWifOK()

    async def test_balance(self):
        await self.assertBalanceOK()

    async def test_balances(self):
        await self.assertBalancesOK()

    async def test_unspent(self):
        await self.assertUnspentOK()

    async def test_unspents(self):
        await self.assertUnspentsOK()

    async def test_merkle_proof(self):
        await self.assertMerkleProofOK()

    async def test_history(self):
        await self.assertHistoryOK()

    async def test_histories(self):
        await self.assertHistoriesOK()

    async def test_block_header(self):
        await self.assertBlockHeaderOK()

    async def test_block_headers(self):
        await self.assertBlockHeadersOK()

    async def test_gettx(self):
        await self.assertGetTXOK()

    async def test_getverbosetx(self):
        await self.assertGetVerboseTXOK()

    async def test_gettxs(self):
        await self.assertTxsOK()

    async def test_balance_merkle_proven(self):
        await self.assertBalanceMerkleProvenOK()

    async def test_balances_merkle_proven(self):
        await self.assertBalancesMerkleProvenOK()

    async def test_transaction(self):
        with mock.patch('cryptos.electrumx_client.client.NotificationSession.send_request',
                        side_effect=self.mock_electrumx_send_request):
            await self.assertTransactionOK(
                "f6e42d4bae97868779ce7377fd34e973c17b38e9029b37e961ae75dd82a9e34f")

    async def test_transaction_multisig(self):
        with mock.patch('cryptos.electrumx_client.client.NotificationSession.send_request',
                        side_effect=self.mock_electrumx_send_request):
            await self.assertMultiSigTransactionOK(
                "d786997a1d0faa3f17b9d597a631822e0e3f29780fbeb4259545f77c40fd7d25")

    async def test_sendmulti_recipient_tx(self):
        with mock.patch('cryptos.electrumx_client.client.NotificationSession.send_request',
                        side_effect=self.mock_electrumx_send_request):
            await self.assertSendMultiRecipientsTXOK("29433a1fed7577398e5414b0c887a59e5a4cfd5a2a26bbd6d5c495bfd46c3061")

    async def test_send(self):
        with mock.patch('cryptos.electrumx_client.client.NotificationSession.send_request',
                        side_effect=self.mock_electrumx_send_request):
            await self.assertSendOK("66b37a63f0bfdc9b9a09543966a7ba22aab9f5c614267ca75ccdc1a85ffcc5a3")

    async def test_subscribe_block_headers(self):
        await self.assertSubscribeBlockHeadersOK()

    async def test_subscribe_block_headers_sync(self):
        await self.assertSubscribeBlockHeadersSyncCallbackOK()

    async def test_latest_block(self):
        await self.assertLatestBlockOK()

    async def test_confirmations(self):
        await self.assertConfirmationsOK()



from .base import BaseCoin


class Litecoin(BaseCoin):
    coin_symbol: str = "LTC"
    display_name: str = "Litecoin"
    segwit_supported: bool = True
    magicbyte: int = 48
    script_magicbyte = 50
    # script_magicbyte: int = 5  # Old magicbyte
    minimum_fee: int = 1000
    wif_prefix: int = 0xb0
    segwit_hrp: str = "ltc"
    hd_path: int = 2
    client_kwargs = {
        'server_file': 'litecoin.json',
    }
    testnet_overrides = {
        'display_name': "Litecoin Testnet",
        'coin_symbol': "LTCTEST",
        'magicbyte': 111,
        'script_magicbyte': 58,
        #'script_magicbyte': 196,  # Old magicbyte,
        'wif_prefix': 0xbf,
        'segwit_hrp': "tltc",
        'minimum_fee': 1000,
        'hd_path': 1,
        'client_kwargs': {
            'server_file': 'litecoin_testnet.json',
            'use_ssl': False
        },
        'xpriv_prefix': 0x04358394,
        'xpub_prefix': 0x043587cf
    }



#!/usr/bin/env python
from setuptools import setup, find_packages

setup(name='cryptos',
      version='2.0.2',
      description='Python Crypto Coin Tools',
      long_description=open('README.md').read(),
      long_description_content_type='text/markdown',
      author='Paul Martin',
      author_email='greatestloginnameever@gmail.com',
      url='http://github.com/primal100/pybitcointools',
      packages=find_packages(),
      include_package_data=True,
      classifiers=[
            'Development Status :: 5 - Production/Stable',
            'Intended Audience :: Developers',
            'Intended Audience :: Education',
            'License :: OSI Approved :: MIT License',
            'Operating System :: OS Independent',
            'Programming Language :: Python',
            'Programming Language :: Python :: 2',
            'Programming Language :: Python :: 3',
            'Topic :: Security :: Cryptography',
      ],
      entry_points='''
            [console_scripts]
            broadcast=scripts.broadcast:main
            convert_private_key=scripts.convert_private_key:main
            create_private_key=scripts.create_private_key:main
            cryptosend=scripts.cryptosend:main
            explorer=scripts.explorer:main
            get_block_sizes=scripts.get_block_sizes:main
            subscribe=scripts.subscribe:main
            view_private_key_addresses=scripts.view_private_key_addresses:main
            '''
      )



import unittest
from queue import Queue
from operator import itemgetter
from cryptos import *
from cryptos import coins
from cryptos.types import Tx, TxOut
from cryptos.electrumx_client.types import ElectrumXTx, ElectrumXMultiBalanceResponse
from typing import Any, Union, List


class BaseSyncCoinTestCase(unittest.TestCase):
    name: str = "Bitcoin Testnet"
    coin = coins.Bitcoin
    addresses: List[str] = ["n2DQVQZiA2tVBDu4fNAjKVBS2RArWhfncv", "mnjBtsvoSo6dMvMaeyfaCCRV4hAF8WA2cu",
                            "mmbKDFPjBatJmZ6pWTW6yqXSC6826YLBX6"]
    segwit_addresses: List[str] = ["2N74sauceDn2qeHFJuNfJ3c1anxPcDRrVtz", "2NDpBxpK4obuGiFodKtYe3dXx14aPwDBPGU",
                                   "2Mt2f4knFtjLZz9CW2979Hw3tYiAYd6WcA1"]
    native_segwit_addresses: List[str] = ["tb1quvys2uxzwl4sqex5xh59kar2y8rt4k7ym0vug3",
                                          "tb1qfuvnn87p787z7nqv9seu4e8fqel83yacg7yf2r",
                                          "tb1qg237zx5qkf0lvweqwnz36969zv4uewapph2pws"]
    multisig_addresses: List[str] = ["2MvmK6SRDc13BaYbumBbtkCH2fKbViC5XEv", "2MtT7kkzRDn1kiT9GZoS1zSgh7twP145Qif"]
    privkeys: List[str] = ["098ddf01ebb71ead01fc52cb4ad1f5cafffb5f2d052dd233b3cad18e255e1db1",
                           "cMrziExc6iMV8vvAML8QX9hGDP8zNhcsKbdS9BqrRa1b4mhKvK6f",
                           "c396c62dfdc529645b822dc4eaa7b9ddc97dd8424de09ca19decce61e6732f71"]
    fee: int = 410
    max_fee: int = 3500
    testnet: bool = True
    min_latest_height: int = 1258030

    unspent_addresses: List[str] = ["ms31HApa3jvv3crqvZ3sJj7tC5TCs61GSA", "2MwHtiGJJqcFgNnbCu1REVy5ooDEeAAFXMy",
                                    "tb1qjap2aae2tsky3ctlh48yltev0sjdmx92yk76wq"]
    unspent: List[ElectrumXTx] = [{'tx_hash': '1d69dd7a23f18d86f514ff7d8ef85894ad00c61fb29f3f7597e9834ac2569c8c',
                                   'tx_pos': 0, 'height': 1238008, 'value': 180000000,
                                   'address': 'ms31HApa3jvv3crqvZ3sJj7tC5TCs61GSA'}]
    unspents: List[ElectrumXTx] = [
        {'tx_hash': '1d69dd7a23f18d86f514ff7d8ef85894ad00c61fb29f3f7597e9834ac2569c8c', 'tx_pos': 0, 'height': 1238008,
         'value': 180000000, 'address': 'ms31HApa3jvv3crqvZ3sJj7tC5TCs61GSA'},
        {'tx_hash': '70bd4ce0e4cf2977ab53e767865da21483977cdb94b1a36eb68d30829c9c392f', 'tx_pos': 1, 'height': 1275633,
         'value': 173980000, 'address': '2MwHtiGJJqcFgNnbCu1REVy5ooDEeAAFXMy'},
        {'tx_hash': '70bd4ce0e4cf2977ab53e767865da21483977cdb94b1a36eb68d30829c9c392f', 'tx_pos': 0, 'height': 1275633,
         'value': 6000000, 'address': 'tb1qjap2aae2tsky3ctlh48yltev0sjdmx92yk76wq'}]
    balance: ElectrumXMultiBalanceResponse = {'confirmed': 180000000, 'unconfirmed': 0}
    balances: List[ElectrumXMultiBalanceResponse] = [
        {'confirmed': 180000000, 'unconfirmed': 0, 'address': 'ms31HApa3jvv3crqvZ3sJj7tC5TCs61GSA'},
        {'confirmed': 173980000, 'unconfirmed': 0, 'address': '2MwHtiGJJqcFgNnbCu1REVy5ooDEeAAFXMy'},
        {'confirmed': 6000000, 'unconfirmed': 0, 'address': 'tb1qjap2aae2tsky3ctlh48yltev0sjdmx92yk76wq'}]
    history: List[ElectrumXTx] = [
        {'tx_hash': '1d69dd7a23f18d86f514ff7d8ef85894ad00c61fb29f3f7597e9834ac2569c8c', 'height': 1238008}]
    histories: List[ElectrumXTx] = [
        {'tx_hash': '1d69dd7a23f18d86f514ff7d8ef85894ad00c61fb29f3f7597e9834ac2569c8c', 'height': 1238008,
         'address': 'ms31HApa3jvv3crqvZ3sJj7tC5TCs61GSA'},
        {'tx_hash': 'e25d8f4036e44159b0364b45867e08ae47a57dda68ba800ba8abe1fb2dc54a40', 'height': 1275633,
         'address': '2MwHtiGJJqcFgNnbCu1REVy5ooDEeAAFXMy'},
        {'tx_hash': '70bd4ce0e4cf2977ab53e767865da21483977cdb94b1a36eb68d30829c9c392f', 'height': 1275633,
         'address': '2MwHtiGJJqcFgNnbCu1REVy5ooDEeAAFXMy'},
        {'tx_hash': '70bd4ce0e4cf2977ab53e767865da21483977cdb94b1a36eb68d30829c9c392f', 'height': 1275633,
         'address': 'tb1qjap2aae2tsky3ctlh48yltev0sjdmx92yk76wq'}]
    txid: str = "1d69dd7a23f18d86f514ff7d8ef85894ad00c61fb29f3f7597e9834ac2569c8c"
    txheight: int = 1238008
    txinputs: List[TxOut] = [
        {'output': '1b8ae7a7a9629bbcbc13339bc29b258122c8d8670c54e6883d35c6a699e23a33:1', 'value': 190453372316}]
    raw_tx: str = "01000000000101333ae299a6c6353d88e6540c67d8c82281259bc29b3313bcbc9b62a9a7e78a1b0100000017160014ffe21a1b058e7f8dedfcc3f1526f82303cff4fc7ffffffff020095ba0a000000001976a9147e585aa1913cf12e9948e90f67188ee9250d555688acfcb92b4d2c00000017a914e223701f10c2a5e7782ef6e10a2560f4c6e968a2870247304402207f2aa4118eee2ef231eab3afcbf6b01b4c1ca3672bd87a3864cf405741bd2c1d02202ab7502cbc50126f68cb2b366e5b3799d3ec0a3359c6a895a730a6891c7bcae10121023c13734016f27089393f9fc79736e4dca1f27725c68e720e1855202f3fbf037e00000000"

    @classmethod
    def setUpClass(cls):
        print('Starting %s sync tests' % cls.name)

    def setUp(self) -> None:
        self._coin = self.coin(testnet=self.testnet)

    def tearDown(self) -> None:
        self._coin.close()

    def assertUnorderedListEqual(self, list1, list2, key):
        list1 = sorted(list1, key=itemgetter(key))
        list2 = sorted(list2, key=itemgetter(key))
        self.assertEqual(list1, list2)

    @property
    def tx(self) -> Tx:
        return deserialize(self.raw_tx)

    def assertBalanceOK(self):
        result = self._coin.get_balance(self.unspent_addresses[0])
        self.assertEqual(self.balance, result)

    def assertBalancesOK(self):
        result = self._coin.get_balances(*self.unspent_addresses)
        self.assertListEqual(self.balances, result)

    def assertBalanceMerkleProvenOK(self):
        result = self._coin.balance_merkle_proven(self.unspent_addresses[0])
        self.assertEqual(self.balance['confirmed'], result)

    def assertBalancesMerkleProvenOK(self):
        balances = [{'address': tx['address'], 'balance': tx['confirmed']} for tx in self.balances]
        result = self._coin.balances_merkle_proven(*self.unspent_addresses)
        self.assertListEqual(balances, result)

    def assertHistoryOK(self):
        result = self._coin.history(self.unspent_addresses[0])
        self.assertEqual(self.history, result)

    def assertHistoriesOK(self):
        result = self._coin.get_histories(*self.unspent_addresses, merkle_proof=True)
        self.assertUnorderedListsEqual(self.histories, result, "tx_hash")

    def assertUnspentOK(self):
        result = self._coin.unspent(self.unspent_addresses[0])
        self.assertEqual(self.unspent, result)

    def assertUnorderedListsEqual(self, expected: List[Any], result: List[Any], order_by: str):
        expected = sorted(expected, key=lambda d: d[order_by])
        result = sorted(result, key=lambda d: d[order_by])
        self.assertListEqual(expected, result)

    def assertUnspentsOK(self):
        unspent_outputs = self._coin.get_unspents(*self.unspent_addresses, merkle_proof=True)
        self.assertUnorderedListsEqual(self.unspents, unspent_outputs, 'tx_hash')

    def assertMixedSegwitTransactionOK(self):

        # Find which of the three addresses currently has the most coins and choose that as the sender
        segwit_max_value = 0
        segwit_sender = self.segwit_addresses[0]
        segwit_from_addr_i = 0
        segwit_unspents = []

        for i, addr in enumerate(self.segwit_addresses):
            addr_unspents = self._coin.unspent(addr)
            value = sum(o['value'] for o in addr_unspents)
            if value > segwit_max_value:
                segwit_max_value = value
                segwit_sender = addr
                segwit_from_addr_i = i
                segwit_unspents = addr_unspents

        regular_max_value = 0
        regular_sender = None
        regular_from_addr_i = 0
        regular_unspents = []

        for i, addr in enumerate(self.addresses):
            addr_unspents = self._coin.unspent(addr)
            value = sum(o['value'] for o in addr_unspents)
            if value > regular_max_value:
                regular_max_value = value
                regular_sender = addr
                regular_from_addr_i = i
                regular_unspents = addr_unspents

        self.assertIn(regular_sender, self.addresses)
        self.assertIn(segwit_sender, self.segwit_addresses)
        unspents = segwit_unspents + regular_unspents
        total_value = segwit_max_value + regular_max_value
        # Arbitrarily set send value, change value, receiver and change address
        send_value = int(total_value * 0.5)

        if segwit_sender == self.segwit_addresses[0]:
            receiver = self.segwit_addresses[1]
        elif segwit_sender == self.segwit_addresses[1]:
            receiver = self.segwit_addresses[2]
        else:
            receiver = self.segwit_addresses[0]

        if regular_sender == self.addresses[0]:
            change_address = self.addresses[1]
        elif regular_sender == self.addresses[1]:
            change_address = self.addresses[2]
        else:
            change_address = self.addresses[0]

        outs = [{'value': send_value, 'address': receiver}]

        # Create the transaction using all available unspents as inputs
        tx = self._coin.mktx_with_change(unspents, outs, change_addr=change_address)

        segwit_privkey = self.privkeys[segwit_from_addr_i]
        regular_privkey = self.privkeys[regular_from_addr_i]

        # Verify that the private key belongs to the sender address for this network
        self.assertEqual(segwit_sender, self._coin.privtop2wpkh_p2sh(segwit_privkey),
                         msg=f"Private key does not belong to address {segwit_sender} on {self._coin.display_name}")
        self.assertEqual(regular_sender, self._coin.privtoaddr(regular_privkey),
                         msg=f"Private key does not belong to address {regular_sender} on {self._coin.display_name}")

        self.assertTrue(self._coin.is_segwit_or_multisig(segwit_sender))
        self.assertFalse(self._coin.is_segwit_or_multisig(regular_sender))

        # Sign each input with the given private keys
        # Try signing one at a time and also with signall and make sure they give the same result

        tx2 = deepcopy(tx)

        for i in range(0, len(segwit_unspents)):
            tx = self._coin.sign(tx, i, segwit_privkey)
        for i in range(len(segwit_unspents), len(unspents)):
            tx = self._coin.sign(tx, i, regular_privkey)

        tx2 = self._coin.signall(tx2, {segwit_sender: segwit_privkey, regular_sender: regular_privkey})

        self.assertDictEqual(tx, dict(tx2))
        self.assertEqual(serialize(tx), serialize(tx2))

        self.assertEqual(tx['locktime'], 0)
        self.assertEqual(tx['version'], 1)
        fee = self._coin.calculate_fee(tx)
        self.assertGreater(fee, 0)
        self.assertLessEqual(fee, self.max_fee)
        self.assertEqual(len(tx['ins']), len(unspents))
        self.assertEqual(len(tx['outs']), 2)
        self.assertEqual(len(tx['witness']), len(unspents))
        self.assertEqual(tx['marker'], 0)
        self.assertEqual(tx['flag'], 1)

        prev_script = None

        for inp in tx['ins']:
            script = inp['script']
            self.assertIsInstance(script, str)
            self.assertIsNot(script, '')

        for o in tx['outs']:
            script = o['script']
            if prev_script:
                self.assertNotEqual(script, prev_script)
            self.assertIsInstance(script, str)
            self.assertNotEqual(script, '')
            prev_script = script

        prev_script_code = None

        for w in tx['witness']:
            script_code = w['scriptCode']
            self.assertIsInstance(script_code, str)
            if script_code:
                self.assertEqual(w['number'], 2)
                if prev_script_code:
                    self.assertNotEqual(script_code, prev_script_code)
            else:
                self.assertEqual(w['number'], 0)
            prev_script_code = script_code

        tx = serialize(tx)
        print(tx)
        # Push the transaction to the network
        result = self._coin.pushtx(tx)
        self.assertTXResultOK(tx, result)

    def assertSegwitTransactionOK(self):

        # Find which of the three addresses currently has the most coins and choose that as the sender
        max_value = 0
        sender = self.segwit_addresses[0]
        from_addr_i = 0
        unspents = []

        for i, addr in enumerate(self.segwit_addresses):
            addr_unspents = self._coin.unspent(addr)
            value = sum(o['value'] for o in addr_unspents)
            if value > max_value:
                max_value = value
                sender = addr
                from_addr_i = i
                unspents = addr_unspents

        # Arbitrarily set send value, receiver and change address
        send_value = int(max_value * 0.1)

        if sender == self.segwit_addresses[0]:
            receiver = self.segwit_addresses[1]
            change_address = self.segwit_addresses[2]
        elif sender == self.segwit_addresses[1]:
            receiver = self.segwit_addresses[2]
            change_address = self.segwit_addresses[0]
        else:
            receiver = self.segwit_addresses[0]
            change_address = self.segwit_addresses[1]

        outs = [{'value': send_value, 'address': receiver}]

        # Create the transaction using all available unspents as inputs
        tx = self._coin.mktx_with_change(unspents, outs, change_addr=change_address)

        privkey = self.privkeys[from_addr_i]

        # Verify that the private key belongs to the sender address for this network
        self.assertEqual(sender, self._coin.privtop2wpkh_p2sh(privkey),
                         msg=f"Private key does not belong to address {sender} on {self._coin.display_name}")

        # Sign each input with the given private key
        tx = self._coin.signall(tx, privkey)

        self.assertEqual(tx['locktime'], 0)
        self.assertEqual(tx['version'], 1)
        fee = self._coin.calculate_fee(tx)
        self.assertGreater(fee, 0)
        self.assertLessEqual(fee, self.fee * 4)
        self.assertEqual(len(tx['ins']), len(unspents))
        self.assertEqual(len(tx['outs']), 2)
        self.assertEqual(len(tx['witness']), len(unspents))
        self.assertEqual(tx['marker'], 0)
        self.assertEqual(tx['flag'], 1)

        prev_script = None

        for inp in tx['ins']:
            script = inp['script']
            if prev_script:
                self.assertEqual(script, prev_script)
            self.assertIsInstance(script, str)
            self.assertIsNot(script, '')
            prev_script = script

        for o in tx['outs']:
            script = o['script']
            if prev_script:
                self.assertNotEqual(script, prev_script)
            self.assertIsInstance(script, str)
            self.assertEqual(len(script), 46)
            prev_script = script

        prev_script_code = None
        for w in tx['witness']:
            self.assertEqual(w['number'], 2)
            script_code = w['scriptCode']
            if prev_script_code:
                self.assertNotEqual(script_code, prev_script_code)
            self.assertIsInstance(script_code, str)
            self.assertNotEqual(script_code, '')
            prev_script_code = script_code

        # Push the transaction to the network

        result = self._coin.pushtx(tx)
        self.assertTXResultOK(tx, result)

    def assertNativeSegwitTransactionOK(self):

        # Find which of the three addresses currently has the most coins and choose that as the sender
        max_value = 0
        sender = self.native_segwit_addresses[0]
        from_addr_i = 0
        unspents = []

        for i, addr in enumerate(self.native_segwit_addresses):
            addr_unspents = self._coin.unspent(addr)
            value = sum(o['value'] for o in addr_unspents)
            if value > max_value:
                max_value = value
                sender = addr
                from_addr_i = i
                unspents = addr_unspents

        # Arbitrarily set send value, change value, receiver and change address
        send_value = int(max_value * 0.1)

        if sender == self.native_segwit_addresses[0]:
            receiver = self.native_segwit_addresses[1]
            change_address = self.native_segwit_addresses[2]
        elif sender == self.native_segwit_addresses[1]:
            receiver = self.native_segwit_addresses[2]
            change_address = self.native_segwit_addresses[0]
        else:
            receiver = self.native_segwit_addresses[0]
            change_address = self.native_segwit_addresses[1]

        outs = [{'value': send_value, 'address': receiver}]

        # Create the transaction using all available unspents as inputs
        tx = self._coin.mktx_with_change(unspents, outs, change_addr=change_address)

        privkey = self.privkeys[from_addr_i]

        # Verify that the private key belongs to the sender address for this network
        self.assertEqual(sender, self._coin.privtosegwitaddress(privkey),
                         msg=f"Private key does not belong to address {sender} on {self._coin.display_name}")

        # Sign each input with the given private key
        tx = self._coin.signall(tx, privkey)

        self.assertEqual(tx['locktime'], 0)
        self.assertEqual(tx['version'], 1)
        fee = self._coin.calculate_fee(tx)
        self.assertGreater(fee, 0)
        self.assertLessEqual(fee, self.fee * 2)
        self.assertEqual(len(tx['ins']), len(unspents))
        self.assertEqual(len(tx['outs']), 2)
        self.assertEqual(len(tx['witness']), len(unspents))
        self.assertEqual(tx['marker'], 0)
        self.assertEqual(tx['flag'], 1)

        prev_script = None

        for inp in tx['ins']:
            script = inp['script']
            self.assertEqual(script, '')

        for o in tx['outs']:
            script = o['script']
            if prev_script:
                self.assertNotEqual(script, prev_script)
            self.assertIsInstance(script, str)
            self.assertEqual(len(script), 44)
            prev_script = script

        prev_script_code = None
        for w in tx['witness']:
            self.assertEqual(w['number'], 2)
            script_code = w['scriptCode']
            if prev_script_code:
                self.assertNotEqual(script_code, prev_script_code)
            self.assertIsInstance(script_code, str)
            self.assertNotEqual(script_code, '')
            prev_script_code = script_code

        tx = serialize(tx)

        # Push the transaction to the network
        result = self._coin.pushtx(tx)
        self.assertTXResultOK(tx, result)

    def assertTXResultOK(self, tx: Union[str, Tx], result):
        if not isinstance(tx, str):
            tx = serialize(tx)
        tx_hash = public_txhash(tx)
        self.assertEqual(result, tx_hash)
        print("TX %s broadcasted successfully" % result)

    def assertTransactionOK(self):

        # Find which of the three addresses currently has the most coins and choose that as the sender
        max_value = 0
        sender = self.addresses[0]
        from_addr_i = 0
        unspents = []

        for i, addr in enumerate(self.addresses):
            addr_unspents = self._coin.unspent(addr)
            value = sum(o['value'] for o in addr_unspents)
            if value > max_value:
                max_value = value
                sender = addr
                from_addr_i = i
                unspents = addr_unspents

        # Arbitrarily set send value, receiver and change address
        send_value = int(max_value * 0.1)

        if sender == self.addresses[0]:
            receiver = self.addresses[1]
            change_address = self.addresses[2]
        elif sender == self.addresses[1]:
            receiver = self.addresses[2]
            change_address = self.addresses[0]
        else:
            receiver = self.addresses[0]
            change_address = self.addresses[1]

        outs = [{'value': send_value, 'address': receiver}]

        # Create the transaction using all available unspents as inputs
        tx = self._coin.mktx_with_change(unspents, outs, change_addr=change_address)

        privkey = self.privkeys[from_addr_i]

        # Verify that the private key belongs to the sender address for this network
        self.assertEqual(sender, self._coin.privtoaddr(privkey),
                         msg=f"Private key does not belong to address {sender} on {self._coin.display_name}")

        # Sign each input with the given private key
        tx = self._coin.signall(tx, privkey)

        self.assertEqual(tx['locktime'], 0)
        self.assertEqual(tx['version'], 1)
        fee = self._coin.calculate_fee(tx)
        self.assertGreater(fee, 0)
        self.assertLessEqual(fee, self.max_fee)
        self.assertEqual(len(tx['ins']), len(unspents))
        self.assertEqual(len(tx['outs']), 2)
        self.assertNotIn('witness', tx)
        self.assertNotIn('marker', tx)
        self.assertNotIn('flag', tx)

        prev_script = None

        for inp in tx['ins']:
            script = inp['script']
            if prev_script:
                self.assertNotEqual(script, prev_script)
            self.assertIsInstance(script, str)
            self.assertIsNot(script, '')
            prev_script = script

        for o in tx['outs']:
            script = o['script']
            if prev_script:
                self.assertNotEqual(script, prev_script)
            self.assertIsInstance(script, str)
            self.assertEqual(len(script), 50)
            prev_script = script

        # Serialize and push the transaction to the network
        tx_serialized = serialize(tx)
        result = self._coin.pushtx(tx)
        self.assertTXResultOK(tx_serialized, result)

    def delete_key_by_name(self, obj, key):
        if isinstance(obj, dict):
            for k, v  in obj.items():
                if k == key:
                    del obj[k]
                    self.delete_key_by_name(obj, key)
                    break
                elif isinstance(v, (dict, list)):
                    self.delete_key_by_name(v, key)
        elif isinstance(obj, list):
            for i in obj:
                self.delete_key_by_name(i, key)

    def assertGetTXOK(self):
        tx = self._coin.get_tx(self.txid)
        self.assertListEqual(list(tx.keys()), ['ins', 'outs', 'version', 'locktime', 'tx_hash'])
        self.assertEqual(tx, self.tx)

    def assertGetSegwitTXOK(self):
        tx = self._coin.get_tx(self.txid)
        self.assertListEqual(list(tx.keys()), ['ins', 'outs', 'version', 'marker', 'flag', 'witness', 'locktime'])
        self.assertEqual(tx, self.tx)

    def assertGetVerboseTXOK(self):
        tx = self._coin.get_verbose_tx(self.txid)
        self.assertListEqual(sorted(tx.keys()),
                             ['blockhash', 'blocktime', 'confirmations', 'hash', 'hex', 'locktime', 'size', 'time',
                              'txid', 'version', 'vin', 'vout', 'vsize', 'weight'])

    def assertGetSegwitTxsOK(self):
        txs = self._coin.get_txs(self.txid)
        self.assertListEqual(list(txs[0].keys()),
                             ['ins', 'outs', 'version', 'marker', 'flag', 'witness', 'locktime'])

    def assertMultiSigTransactionOK(self):
        pubs = [privtopub(priv) for priv in self.privkeys]
        script, address1 = self._coin.mk_multsig_address(*pubs, num_required=2)
        self.assertEqual(address1, self.multisig_addresses[0])
        pubs2 = [privtopub(priv) for priv in self.privkeys[0:2]]
        script2, address2 = self._coin.mk_multsig_address(*pubs2)
        self.assertEqual(address2, self.multisig_addresses[1])

        # Find which of the three addresses currently has the most coins and choose that as the sender
        max_value = 0
        sender = None

        for i, addr in enumerate(self.multisig_addresses):
            addr_unspents = self._coin.unspent(addr)
            value = sum(o['value'] for o in addr_unspents)
            if value > max_value:
                max_value = value
                sender = addr

        # Arbitrarily set send value, receiver and change address
        send_value = int(max_value * 0.1)

        receiver = address2 if sender == address1 else address1

        tx = self._coin.preparetx(sender, receiver, send_value, self.fee)

        for i in range(0, len(tx['ins'])):
            if sender == address1:
                sig1 = self._coin.multisign(tx, i, script, self.privkeys[0])
                sig3 = self._coin.multisign(tx, i, script, self.privkeys[2])
                tx = apply_multisignatures(tx, i, script, sig1, sig3)
            else:
                sig1 = self._coin.multisign(tx, i, script, self.privkeys[0])
                sig2 = self._coin.multisign(tx, i, script, self.privkeys[1])
                tx = apply_multisignatures(tx, i, script, sig1, sig2)

        self.assertEqual(tx['locktime'], 0)
        self.assertEqual(tx['version'], 1)
        fee = self._coin.calculate_fee(tx)
        self.assertEqual(fee, self.fee)
        self.assertGreaterEqual(len(tx['ins']), 1)
        self.assertEqual(len(tx['outs']), 2)
        self.assertNotIn('witness', tx)
        self.assertNotIn('marker', tx)
        self.assertNotIn('flag', tx)

        prev_script = None

        for inp in tx['ins']:
            script = inp['script']
            if prev_script:
                self.assertNotEqual(script, prev_script)
            self.assertIsInstance(script, str)
            self.assertIsNot(script, '')
            prev_script = script

        for o in tx['outs']:
            script = o['script']
            if prev_script:
                self.assertNotEqual(script, prev_script)
            self.assertIsInstance(script, str)
            self.assertNotEqual(script, '')
            prev_script = script

        # Push the transaction to the network
        result = self._coin.pushtx(tx)
        print(serialize(tx))
        self.assertTXResultOK(tx, result)

    def assertBlockHeaderOK(self):
        blockinfo = self._coin.block_header(self.txheight)
        self.assertListEqual(sorted(blockinfo.keys()),
                             ['bits', 'hash', 'merkle_root', 'nonce', 'prevhash', 'timestamp', 'version']
                             )

    def assertBlockHeadersOK(self):
        blockinfos = self._coin.block_headers(self.txheight)
        for blockinfo in blockinfos:
            self.assertListEqual(sorted(blockinfo.keys()),
            ['bits', 'hash', 'merkle_root', 'nonce', 'prevhash', 'timestamp', 'version']
        )

    def assertMerkleProofOK(self):
        tx = self.unspent[0]
        tx_hash = tx['tx_hash']
        proof = self._coin.merkle_prove(self.unspent[0])
        self.assertDictEqual(dict(proof), {
            'tx_hash': tx_hash,
            'proven': True
        })

    def assertSendMultiRecipientsTXOK(self):

        # Find which of the three addresses currently has the most coins and choose that as the sender
        max_value = 0
        sender = self.addresses[0]
        from_addr_i = 0

        for i, addr in enumerate(self.addresses):
            addr_unspents = self._coin.unspent(addr)
            value = sum(o['value'] for o in addr_unspents)
            if value > max_value:
                max_value = value
                sender = addr
                from_addr_i = i

        privkey = self.privkeys[from_addr_i]

        # Arbitrarily set send value, change value, receiver and change address
        outputs_value = max_value - self.fee
        send_value1 = int(outputs_value * 0.2)
        send_value2 = int(outputs_value * 0.5)

        if sender == self.addresses[0]:
            receiver1 = self.addresses[1]
            receiver2 = self.addresses[2]
        elif sender == self.addresses[1]:
            receiver1 = self.addresses[2]
            receiver2 = self.addresses[0]
        else:
            receiver1 = self.addresses[0]
            receiver2 = self.addresses[1]

        self.assertEqual(sender, self._coin.privtoaddr(privkey),
                         msg=f"Private key does not belong to address {sender} on {self._coin.display_name}")

        outs = [{'address': receiver1, 'value': send_value1},  {'address': receiver2, 'value': send_value2}]

        result = self._coin.send_to_multiple_receivers_tx(privkey, sender, outs, fee=self.fee)
        self.assertIsInstance(result, str)
        print("TX %s broadcasted successfully" % result)

    def assertSendOK(self):

        # Find which of the three addresses currently has the most coins and choose that as the sender
        max_value = 0
        sender = self.addresses[0]
        from_addr_i = 0

        for i, addr in enumerate(self.addresses):
            addr_unspents = self._coin.unspent(addr)
            value = sum(o['value'] for o in addr_unspents)
            if value > max_value:
                max_value = value
                sender = addr
                from_addr_i = i
                break

        privkey = self.privkeys[from_addr_i]

        self.assertEqual(sender, self._coin.privtoaddr(privkey),
                         msg=f"Private key does not belong to address {sender} on {self._coin.display_name}")

        # Arbitrarily set send value, change value, receiver and change address
        send_value = int(max_value * 0.2)

        if sender == self.addresses[0]:
            receiver = self.addresses[1]
        elif sender == self.addresses[1]:
            receiver = self.addresses[2]
        else:
            receiver = self.addresses[0]

        result = self._coin.send(privkey, sender, receiver, send_value)
        self.assertIsInstance(result, str)
        print("TX %s broadcasted successfully" % result)

    def assertSubscribeBlockHeadersOK(self):
        queue = Queue()
        block_keys = ['bits', 'hash', 'merkle_root', 'nonce', 'prevhash', 'timestamp', 'version']

        def on_new_block(height: int, hex_header: str, header: BlockHeader) -> None:
            queue.put((height, hex_header, header))

        self._coin.subscribe_to_block_headers(on_new_block)
        result = queue.get()
        height, hex_header, header = result
        self.assertGreater(height, self.min_latest_height)
        self.assertEqual(deserialize_header(binascii.unhexlify(hex_header)), header)
        self.assertListEqual(sorted(header.keys()), block_keys)
        self._coin.unsubscribe_from_block_headers()

    def assertLatestBlockOK(self):
        height, hex_header, header = self._coin.block
        self.assertGreater(height, self.min_latest_height)
        self.assertIsInstance(hex_header, str)
        self.assertIsInstance(header, dict)
        height, hex_header, header = self._coin.block
        self.assertGreater(height, self.min_latest_height)
        self.assertIsInstance(hex_header, str)
        self.assertIsInstance(header, dict)

    def assertConfirmationsOK(self):
        confirmations = self._coin.confirmations(0)
        block = self._coin.block
        height = block[0]
        self.assertEqual(confirmations, 0)
        confirmations = self._coin.confirmations(height - 1)
        self.assertEqual(confirmations, 2)
        confirmations = self._coin.confirmations(1)
        self.assertEqual(confirmations, height)

    def assertSubscribeAddressOK(self):
        queue = Queue()
        address = self.addresses[0]

        def add_to_queue(addr: str, status: str) -> None:
            queue.put((addr, status))

        self._coin.subscribe_to_address(add_to_queue, address)
        addr, initial_status = queue.get()
        self.assertEqual(addr, address)
        self.assertTransactionOK()
        addr, status = queue.get()
        self.assertEqual(addr, address)
        self.assertNotEqual(initial_status, status)
        self._coin.unsubscribe_from_address(address)

    def assertSubscribeAddressTransactionsOK(self):
        queue = Queue()
        address = self.addresses[0]

        def add_to_queue(address: str, txs: List[Tx], newly_confirmed: List[Tx], history: List[Tx],
                         unspent: List[Tx], confirmed: int, unconfirmed: int, proven: int) -> None:
            queue.put((address, txs, newly_confirmed, history, unspent, confirmed, unconfirmed, proven))

        self._coin.subscribe_to_address_transactions(add_to_queue, address)
        addr, start_txs, start_newly_confirmed, start_history, start_unspent, start_confirmed, start_unconfirmed, start_proven = queue.get()
        self.assertEqual(addr, address)
        self.assertEqual(start_txs, [])
        self.assertEqual(start_newly_confirmed, [])
        self.assertSendOK()
        addr, new_txs, newly_confirmed, history, unspent, confirmed, unconfirmed, proven = queue.get()
        self.assertEqual(addr, address)
        self.assertGreaterEqual(len(new_txs), 1)
        self.assertEqual(len(newly_confirmed), 0)
        self.assertGreaterEqual(len(history), 9)
        self.assertEqual(len(start_history), len(history))
        self.assertNotEqual(unconfirmed, start_unconfirmed)
        self._coin.unsubscribe_from_address(address)




from .main import *
from .transaction import select

class Wallet:
    def __init__(self, keystore, transaction_history=None):
        self.coin = keystore.coin
        self.keystore = keystore
        self.address_derivations = {}
        self.is_watching_only = self.keystore.is_watching_only()
        self.transaction_history = transaction_history or []
        self.xtype = self.keystore.xtype
        if self.keystore.electrum:
            self.script_type = self.keystore.xtype
        else:
            self.script_type = "p2pkh"

    def privkey(self, address, formt="wif_compressed", password=None):
        if self.is_watching_only:
            return
        try:
            addr_derivation = self.address_derivations[address]
        except KeyError:
            raise Exception(
                "Address %s has not been generated yet. Generate new addresses with new_receiving_addresses or new_change_addresses methods" % address)
        pk, compressed = self.keystore.get_private_key(addr_derivation, password)
        return self.coin.encode_privkey(pk, formt, script_type=self.script_type)

    def export_privkeys(self, password=None):
        if self.is_watching_only:
            return
        return {
            'receiving': {addr: self.privkey(addr, password=password) for addr in self.receiving_addresses},
            'change': {addr: self.privkey(addr, password=password) for addr in self.change_addresses}
        }

    def receiving_address(self, index):
        pubkey = self.keystore.keypairs.keys()(index)
        address = self.pubtoaddr(pubkey)
        self.address_derivations[address] = pubkey
        return address

    def change_address(self, index):
        pubkey =  self.keystore.keypairs.keys()(index)
        address = self.pubtoaddr(pubkey)
        self.address_derivations[address] = pubkey
        return address

    def pubtoaddr(self, pubkey):
        if self.xtype == "p2pkh":
            return self.coin.pubtoaddr(pubkey)
        elif self.xtype == "p2wpkh":
            return self.coin.pub_to_segwit_address(pubkey)
        elif self.xtype == "p2wpkh-p2sh":
            return self.coin.pubtop2wpkh_p2sh(pubkey)

    @property
    def addresses(self):
        return [self.pubtoaddr(pub) for pub in self.keystore.keypairs.keys()]

    @property
    def receiving_addresses(self):
        return self.addresses

    @property
    def change_addresses(self):
        return self.addresses

    def select_receive_address(self):
        return self.addresses[0]

    def select_change_address(self):
        return self.addresses[0]

    def new_receiving_address_range(self, num):
        return self.receiving_addresses[0]

    def new_change_address_range(self, num):
        return self.receiving_addresses[0]

    def new_receiving_addresses(self, num=10):
        return self.addresses

    def new_change_addresses(self, num=10):
        return self.addresses

    def new_receiving_address(self):
        return self.new_receiving_addresses(num=1)[0]

    def new_change_address(self):
        return self.new_change_addresses(num=1)[0]

    def is_mine(self, address):
        return address in self.addresses

    def is_change(self, address):
        return True

    def get_balances(self):
        return self.coin.get_balance(*self.addresses)

    def balance(self):
        balances = self.get_balances()
        confirmed_balance = sum(b['confirmed'] for b in balances)
        unconfirmed_balance = sum(b['unconfirmed'] for b in balances)
        return {
            'total': confirmed_balance + unconfirmed_balance,
            'unconfirmed': unconfirmed_balance,
            'confirmed': confirmed_balance
        }

    def unspent(self, addresses=None, merkle_proof=False):
        addresses = addresses or self.addresses
        return self.coin.unspent(*addresses, merkle_proof=merkle_proof)

    def select_unspents(self, value, addresses=None, merkle_proof=False):
        unspents = self.unspent(addresses=addresses, merkle_proof=merkle_proof)
        return select(unspents, value)

    def history(self, addresses=None, merkle_proof=False):
        addresses = addresses or self.addresses
        return self.coin.history(*addresses, merkle_proof=merkle_proof)

    def synchronise(self):
        tx_hashes = [tx['tx_hash'] for tx in self.transaction_history]
        txs = self.history()
        new_txs = [tx for tx in txs if tx['tx_hash'] not in tx_hashes]
        self.transaction_history += self.coin.filter_by_proof(*new_txs)

    def sign(self, txobj, password=None):
        if self.is_watching_only:
            return
        pkeys_for = [inp['address'] for inp in txobj['ins']]
        privkeys = {address: self.privkey('address', password) for address in pkeys_for}
        return self.coin.signall(txobj, privkeys)

    def pushtx(self, tx_hex):
        return self.coin.pushtx(tx_hex)

    def preparemultitx(self, outs, fee=50000, change_addr=None, fee_for_blocks=0, addresses=None):
        change = change_addr or self.select_change_address()
        value = sum(out['value'] for out in outs) + fee
        ins = self.select_unspents(value, addresses=addresses)
        if self.coin.segwit_supported:
            if self.xtype == 'p2pkh':
                for i in ins:
                    i['segwit'] = False
                    i['new_segwit'] = False
            elif self.xtype == "p2wpkh-p2sh":
                for i in ins:
                    i['segwit'] = True
                    i['new_segwit'] = False
            elif self.xtype == 'p2wpkh':
                for i in ins:
                    i['segwit'] = True
                    i['new_segwit'] = True
        return self.coin.mktx_with_change(ins, outs, fee=fee, fee_for_blocks=fee_for_blocks, change=change)

    def preparetx(self, to, value, fee=50000,  fee_for_blocks=0, change_addr=None, addresses=None):
        outs = [{'address': to, 'value': value}]
        return self.preparemultitx(outs, fee=fee, fee_for_blocks=fee_for_blocks, change_addr=change_addr,
                                   addresses=addresses)

    def preparesignedtx(self, to, value, fee=50000, fee_for_blocks=0, change_addr=None, addresses=addresses,
                        password=None):
        txobj = self.preparetx(to, value, fee=fee, fee_for_blocks=fee_for_blocks, change_addr=change_addr,
                               addresses=addresses)
        return self.sign(txobj, password=password)

    def preparesignedmultitx(self, outs, fee=50000, fee_for_blocks=0, change_addr=None, addresses=None, password=None):
        txobj = self.preparemultitx(outs, fee=fee, change_addr=change_addr, addresses=addresses,
                                    fee_for_blocks=fee_for_blocks)
        return self.sign(txobj, password=password)

    def send(self, to, value, fee=50000, fee_for_blocks=0, change_addr=None, addresses=None, password=None):
        tx = self.preparesignedtx(to, value, fee=fee, fee_for_blocks=fee_for_blocks, change_addr=change_addr,
                                  addresses=addresses, password=password)
        return self.pushtx(tx)

    def sendmultitx(self, outs, fee=50000, fee_for_blocks=0, change_addr=None, addresses=None, password=None):
        tx = self.preparesignedmultitx(outs, fee=fee, fee_for_blocks=fee_for_blocks, change_addr=change_addr,
                                       addresses=addresses, password=password)
        return self.pushtx(tx)

class HDWallet(Wallet):
    def __init__(self, keystore, transaction_history=None, num_addresses=0, last_receiving_index=0, last_change_index=0):
        super(HDWallet, self).__init__(keystore, transaction_history=transaction_history)
        self.last_receiving_index = last_receiving_index
        self.last_change_index = last_change_index
        self.new_receiving_addresses(num=num_addresses)
        self.new_change_addresses(num=num_addresses)
        self.used_addresses = self.get_used_addresses()
        self.xtype = self.keystore.xtype
        if self.keystore.electrum:
            self.script_type = self.keystore.xtype
        else:
            self.script_type = "p2pkh"

    def privkey(self, address, formt="wif_compressed", password=None):
        if self.is_watching_only:
            return
        try:
            addr_derivation = self.address_derivations[address]
        except KeyError:
            raise Exception(
                "Address %s has not been generated yet. Generate new address_derivations with new_receiving_addresses or new_change_addresses methods" % address)
        pk, compressed = self.keystore.get_private_key(addr_derivation, password)
        return self.coin.encode_privkey(pk, formt, script_type=self.script_type)

    def export_privkeys(self, password=None):
        if self.is_watching_only:
            return
        return {
            'receiving': {addr: self.privkey(addr, password=password) for addr in self.receiving_addresses},
            'change': {addr: self.privkey(addr, password=password) for addr in self.change_addresses}
        }

    def pubkey_receiving(self, index):
        return self.keystore.derive_pubkey(0, index)

    def pubkey_change(self, index):
        return self.keystore.derive_pubkey(1, index)

    def pubtoaddr(self, pubkey):
        if self.xtype == "p2pkh":
            return self.coin.pubtoaddr(pubkey)
        elif self.xtype == "p2wpkh":
            return self.coin.pub_to_segwit_address(pubkey)
        elif self.xtype == "p2wpkh-p2sh":
            return self.coin.pubtop2wpkh_p2sh(pubkey)

    def receiving_address(self, index):
        pubkey = self.pubkey_receiving(index)
        address = self.pubtoaddr(pubkey)
        self.address_derivations[address] = (0, index)
        return address

    def change_address(self, index):
        pubkey = self.pubkey_change(index)
        address = self.pubtoaddr(pubkey)
        self.address_derivations[address] = (1, index)
        return address

    @property
    def addresses(self):
        return self.address_derivations.keys()

    @property
    def receiving_addresses(self):
        return [addr for addr in self.address_derivations.keys() if not self.address_derivations[addr][0]]

    @property
    def change_addresses(self):
        return [addr for addr in self.address_derivations.keys() if self.address_derivations[addr][0]]

    def new_receiving_address_range(self, num):
        index = self.last_receiving_index
        return range(index, index+num)

    def new_change_address_range(self, num):
        index = self.last_change_index
        return range(index, index+num)

    def new_receiving_addresses(self, num=10):
        addresses = list(map(self.receiving_address, self.new_receiving_address_range(num)))
        self.last_receiving_index += num
        return addresses

    def new_change_addresses(self, num=10):
        addresses = list(map(self.change_address, self.new_change_address_range(num)))
        self.last_change_index += num
        return addresses

    def new_receiving_address(self):
        return self.new_receiving_addresses(num=1)[0]

    def new_change_address(self):
        return self.new_change_addresses(num=1)[0]

    def select_receive_address(self):
        try:
            return next(addr for addr in self.receiving_addresses if addr not in self.used_addresses)
        except StopIteration:
            return self.new_receiving_address()

    def select_change_address(self):
        try:
            return next(addr for addr in self.receiving_addresses if addr not in self.used_addresses)
        except StopIteration:
            return self.new_change_address()

    def is_change(self, address):
        return address in self.change_addresses

    def get_used_addresses(self):
        return list(set([tx['addr'] for tx in self.transaction_history]))

    def synchronise(self):
        super(HDWallet, self).synchronise()
        self.used_addresses = self.get_used_addresses()
